import logging
import json
from datamaestro.download import Resource
from typing import Callable, Iterator
from pathlib import Path
import requests
import random
import re
from requests.exceptions import HTTPError
from tqdm.auto import tqdm
import time
import urllib.parse
import uuid


wayback_prefix = re.compile(r"^https:\/\/web\.archive\.org\/web")
replace_pattern = re.compile(r"(web\.archive\.org\/web\/\d+)")


def download_with_retry(url: str, max_retries: int = 10) -> requests.Response:
    """Download a URL with exponential backoff, until max_retries is reached."""
    retry_num = 0
    while True:
        try:
            response = requests.get(url)
            response.raise_for_status()
            return response
        except HTTPError as e:
            status_code = e.response.status_code
            if not (status_code == 429 or status_code >= 500):
                # This is not an error we should retry on
                raise e

            if retry_num > max_retries:
                logging.error(
                    f"Failed to perform GET request on {url}"
                    f"after {max_retries} retries."
                )
                raise e

            if status_code == 429:
                time.sleep(5 + 2**retry_num + random.randint(0, 1000) / 1000)
            else:
                time.sleep(2**retry_num + random.randint(0, 1000) / 1000)
            retry_num += 1


def download_link(link: str, timestamp: str):
    page_id = str(uuid.uuid4())
    url_no_header = None

    try:
        # Find the Wayback Machine link
        if not wayback_prefix.match(link):
            link_encoded = urllib.parse.quote(link)

            available, availability_attempt = False, 0
            # Sometimes the API returns HTTP success code 200, but archived
            # snapshots shows page is unavailable when it actually is. Give it a
            # total of three tries.
            while not available and availability_attempt < 3:
                response = download_with_retry(
                    "http://archive.org/wayback/available?"
                    f"url={link_encoded}&timestamp={timestamp}"
                )
                json_response = response.json()
                available = "closest" in json_response["archived_snapshots"]
                availability_attempt += 1

            if not available:
                logging.warning(
                    f"Not available on Wayback Machine: {link}, "
                    f"HTTP code {response.status_code}, {json_response}"
                )
                return {"link": link, "page_id": page_id, "available": False}

            url = json_response["archived_snapshots"]["closest"]["url"]
        else:
            url = link

        match = replace_pattern.search(url)
        assert match
        url_no_header = replace_pattern.sub(f"{match.group(1)}id_", url)

        response = download_with_retry(url_no_header)
        html_page = response.text

        return {
            "link": link,
            "id": url_no_header,
            "contents": html_page,
        }

    except HTTPError as http_err:
        logging.warning(f"HTTP error occurred: {http_err} for {link}")
        return {
            "link": link,
            "page_id": page_id,
            "available": False,
            "status_code": http_err.response.status_code if http_err.response else None,
            "wayback_url": url_no_header,
        }
    except UnicodeDecodeError as e:
        logging.warning(f"Unicode decode error occurred: {e} for {link}")
        return {
            "link": link,
            "page_id": page_id,
            "available": False,
            "status_code": response.status_code,
            "wayback_url": url_no_header,
        }
    except Exception as e:
        logging.warning(f"Exception occurred: {e} for {link}")
        return {
            "link": link,
            "page_id": page_id,
            "available": False,
            "status_code": None,
            "wayback_url": url_no_header,
        }


class wayback_documents(Resource):
    """Collect documents from wayback"""

    def __init__(self, timestamp: str, urls_fn: Callable[[], Iterator[str]], name=None):
        super().__init__(name)
        self.timestamp = timestamp
        self.urls_fn = urls_fn

    def prepare(self):
        return self.dataset.datapath / self.name

    def download(self, force=False):
        # Creates directory if needed
        destination: Path = self.dataset.datapath / self.name
        self.dataset.datapath.mkdir(exist_ok=True)

        # Early exit
        done_path = destination.with_suffix(".done")
        if done_path.is_file() and not force:
            return True

        # Reads the URLs
        logging.info("Retrieving URLs from wayback into %s", destination)
        pos = 0
        urls = set()
        with destination.open("at+") as fp:
            fp.seek(0)
            try:
                while line := fp.readline():
                    pos = fp.tell()
                    urls.add(json.loads(line)["url"])
            except json.JSONDecodeError:
                logging.warning(f"JSON decoding error: getting back to position {pos}")
                fp.seek(pos)

            # Get the remaining ones
            for url in tqdm(self.urls_fn()):
                if url not in urls:
                    fp.write(json.dumps(download_link(url, self.timestamp)))

        # Everything is fine
        done_path.touch()



from inspect import ismodule
import logging
import io
import re
from pathlib import Path
import inspect
import typing
from typing import Dict, Iterable, Optional
import importlib
from datamaestro.definitions import AbstractDataset

import mkdocs
import mkdocs.config
import mkdocs.plugins
from mkdocs.structure.files import File as MkdocFile
from mkdocs.structure.pages import Page as MkdocPage

from docstring_parser import parse as docstring_parse

import experimaestro
import experimaestro.mkdocs.base
from experimaestro.core.types import ObjectType

from ..context import Context, Repository, Datasets


# Custom URIs for tags and modules
RE_MODULE = re.compile(r"^datamaestro/df/([^/]*)/(.*)\.md$")
RE_TASK = re.compile(r"^datamaestro/task/([^/]*)\.md$")
RE_APIGEN = re.compile(r"@@api:(.+)")


class Matcher:
    def __init__(self):
        self.match = None

    def __call__(self, re, value):
        self.match = re.match(value)
        return self.match

    def group(self, *args):
        return self.match.group(*args)


MATCHER = Matcher()


class ClassificationItem:
    def __init__(self, name):
        self.values = []
        self.name = name


def method_documentation(doc, types):
    s = ""
    for param in doc.params:
        type_name = param.type_name
        if not type_name and param.arg_name in types:
            t = types[param.arg_name]
            if isinstance(t, typing._Final):
                type_name = str(t).replace("typing.", "")
            else:
                type_name = types[param.arg_name].__name__
        s += " - `{}` (`{}`): {}\n".format(
            param.arg_name, type_name or "?", param.description or ""
        )
    return s


def document_data(datatype: ObjectType):
    xpm = datatype.__getxpmtype__()
    s = "### %s\n\nClass `%s.%s`\n\n" % (
        xpm.identifier,
        datatype.__module__,
        datatype.__name__,
    )

    supertypes = ", ".join(str(parent.identifier) for parent in xpm.parents())
    if supertypes:
        s += "**Supertypes**: %s\n\n" % supertypes

    arguments = list(xpm.arguments.items())
    if arguments:
        s += "\n**Arguments**\n\n"
        for name, argument in arguments:
            s += "- **%s**: %s\n" % (name, argument.help)

    for name, method in inspect.getmembers(datatype):
        if inspect.isfunction(method) and hasattr(method, "__datamaestro_doc__"):
            try:
                doc = docstring_parse(method.__datamaestro_doc__)

                signature = re.sub(r"\(self,?", "(", str(inspect.signature(method)))
                s += "#### %s%s\n" % (name, signature)

                if doc.short_description:
                    s += doc.short_description + "\n"
                if doc.long_description:
                    s += doc.long_description + "\n"
                s += method_documentation(doc, method.__annotations__)
            except Exception:
                logging.error(
                    "Error while parsing documetnation of %s (%s)",
                    method,
                    method.__module__,
                )

    return s


def document_object(object):
    try:
        name = object.__name__
        # Get the documentation
        if inspect.isclass(object):
            if hasattr(object, "__datamaestro__"):
                return document_data(object)

            docstring = object.__init__.__doc__
            types = object.__init__.__annotations__
            signature = str(inspect.signature(object.__init__)).replace("(self, ", "(")

        else:
            docstring = object.__doc__
            types = object.__annotations__
            signature = str(inspect.signature(object))

        doc = docstring_parse(docstring)
        if doc.short_description:
            s = "### " + doc.short_description + "\n\n"
        else:
            s = "### %s\n\n" % name

        s += "`@{}{}`\n\n".format(name, signature)

        if doc.long_description:
            s += doc.long_description

        s += method_documentation(doc, types)

        return s

    except Exception:
        logging.exception(
            "Exception while generating the documentation for %s" % object.__name__
        )
        return r"""<div class="error">Documentation error for {}</div>""".format(
            object.__name__
        )


def document(match):
    """Generate the documentation"""
    from datamaestro.data import Base

    modulename, name = match.group(1).rsplit(".", 1)

    module = importlib.import_module(modulename)
    try:
        object = getattr(module, name)
    except Exception:
        return "<div class='error'>Cannot find %s in %s</div>" % (
            name,
            modulename,
        )

    if ismodule(object):
        return "\n\n".join(
            document_data(o)
            for o in object.__dir__()
            if inspect.isclass(o) and hasattr(Base, "__datamaestro__")
        )
    return document_object(object)


class Classification:
    def __init__(self, name):
        self.id = name.lower()
        self.name = name
        # Maps keys to couple
        self.map = {}

        self.re = re.compile(r"^datamaestro/%s/([^/]*)\.md$" % (self.id))

    def add(self, name, value):
        key = name.lower()
        if key not in self.map:
            self.map[key] = ClassificationItem(name)
        self.map[key].values.append(value)

    def addFiles(self, files, config):
        files.append(
            MkdocFile("datamaestro/%s.md" % self.id, "", config["site_dir"], False)
        )
        for key in self.map.keys():
            files.append(
                MkdocFile(
                    "datamaestro/%s/%s.md" % (self.id, key),
                    "",
                    config["site_dir"],
                    False,
                )
            )

    def match(self, path):
        if path == "datamaestro/%s.md" % self.id:
            r = io.StringIO()
            r.write("# List of %s\n\n" % self.name)
            for key, value in sorted(self.map.items(), key=lambda kv: kv[0]):
                r.write("- [%s](%s/%s.html)\n" % (value.name, self.id, key))
            return r.getvalue()

        if MATCHER(self.re, path):
            # Case of a tag
            r = io.StringIO()
            key = MATCHER.group(1)
            item = self.map[key]
            r.write("# %s\n\n" % item.name)

            for ds in item.values:
                meta = ds

                module = Datasets(importlib.import_module(meta.t.__module__))
                r.write(
                    "- [%s](../df/%s/%s.html#%s)\n"
                    % (
                        meta.name or meta.id,
                        meta.repository.id,
                        module.id,
                        meta.id,
                    )
                )

            return r.getvalue()

    @property
    def nav(self):
        nav = [{"List of %s" % self.id: "datamaestro/%s.md" % self.id}]
        for key, item in self.map.items():
            nav.append({item.name: "datamaestro/%s/%s.md" % (self.id, key)})
        return nav


class DatasetGenerator(mkdocs.plugins.BasePlugin):
    """Mkdocs plugin for datamaestro submodules

    See:
        https://www.mkdocs.org/user-guide/plugins/ for mkdocs plugins

    Arguments:
        mkdocs {[type]} -- [description]
    """

    CONF: Optional[Context] = None
    REPOSITORY: Optional[Repository] = None

    config_scheme = (("repository", mkdocs.config.config_options.Type(str)),)
    experimaestro_plugin = experimaestro.mkdocs.base.Documentation

    def on_pre_build(self, config):
        self.experimaestro_plugin = config["plugins"].get("experimaestro")
        if self.experimaestro_plugin is None:
            logging.warning(
                "No experimaestro plugin found: won't generate type information"
            )

    @staticmethod
    def configuration() -> Context:
        if DatasetGenerator.CONF is None:
            DatasetGenerator.CONF = Context()
        return DatasetGenerator.CONF

    @property
    def repository(self) -> Repository:
        if DatasetGenerator.REPOSITORY is None:
            DatasetGenerator.REPOSITORY = DatasetGenerator.configuration().repository(
                self.repository_id
            )
        return DatasetGenerator.REPOSITORY

    def parse_nav(self, nav):
        for entry in nav:
            assert len(entry) == 1
            _, value = *entry.keys(), *entry.values()
            if isinstance(value, list):
                for value in self.parse_nav(value):
                    yield value
            else:
                yield value

    def on_config(self, config):
        self.repository_id = self.config["repository"]
        self.classifications = []
        self.modules: Dict[str, Iterable[AbstractDataset]] = {}

        if not self.repository_id:
            return

        config["extra_css"].insert(0, "mainstyle.css")
        # Navigation
        nav = config["nav"]

        self.tags = Classification("Tags")
        self.tasks = Classification("Tasks")
        self.classifications = [self.tags, self.tasks]

        navdf = []
        nav.append({"Datasets": navdf})

        for module in self.repository.modules():
            if not module.id:
                continue
            path = "datamaestro/df/%s/%s.md" % (self.repository_id, module.id)
            hasdataset = False
            for dataset in module:
                hasdataset = True
                for tag in dataset.tags:
                    self.tags.add(tag, dataset)
                for task in dataset.tasks:
                    self.tasks.add(task, dataset)

            if hasdataset:
                self.modules[module.id] = module
                navdf.append({module.id: path})

        for c in self.classifications:
            nav.append({c.name: c.nav})

        return config

    def on_post_build(self, config):
        """Called after the build"""
        import importlib.resources
        import shutil

        path = Path(config["site_dir"]) / "mainstyle.css"
        with (
            importlib.resources.open_binary(
                "datamaestro.commands", "mainstyle.css"
            ) as source,
            path.open("wb") as dest,
        ):
            shutil.copyfileobj(source, dest)

    def on_files(self, files, config):
        """Called when list of files has been read"""
        if self.repository_id:
            files.append(
                MkdocFile("datamaestro/tasks.md", "", config["site_dir"], False)
            )
            for c in self.classifications:
                c.addFiles(files, config)

            for module in self.modules.values():
                # Add a file for each dataset
                f = MkdocFile(
                    "datamaestro/df/%s/%s.md" % (self.repository_id, module.id),
                    "",
                    config["site_dir"],
                    False,
                )
                files.append(f)
        return files

    def on_serve(self, server, config, builder):
        """Refresh when changing source code"""
        import datamaestro

        path = (
            self.repository.basedir
            if self.repository
            else str(Path(datamaestro.__file__).parent)
        )
        basemodule = (
            "%s." % self.repository.module if self.repository else "datamaestro."
        )

        def rebuild():
            import sys

            # Clear-up loaded module
            toremove = [
                module for module in sys.modules if module.startswith(basemodule)
            ]
            for module in toremove:
                del sys.modules[module]

            # Remove defined data
            ObjectType.REGISTERED = {}

            builder()

        logging.info("Watching %s", path)
        # server.watch(path, rebuild)

    def on_page_markdown(self, markdown, page, config, **kwargs):
        if page.url.startswith("api/"):
            return RE_APIGEN.sub(document, markdown)
        if self.repository and page.url == "":
            return (
                "Documentation for datamaestro module **%s (version %s)**\n\n"
                % (self.repository.NAMESPACE, self.repository.version())
            ) + markdown
        return markdown

    def on_page_read_source(self, page: MkdocPage, **kwargs):
        """Generate pages"""
        path = page.file.src_path

        # --- Classifications
        for c in self.classifications:
            r = c.match(path)
            if r:
                return r

        # --- Dataset file documentation generation

        m = RE_MODULE.match(path)
        if not m:
            return None

        df = self.modules[m.group(2)]
        r = io.StringIO()

        r.write("# %s\n" % df.id)
        r.write(df.description)

        r.write("\n\n")
        r.write("## List of datasets\n\n")
        for ds in df:
            r.write(
                (
                    """<div class="dataset-entry"><div class='dataset-id'>"""
                    """%s<a name="%s"></a></div>\n\n"""
                )
                % (ds.id, ds.id)
            )
            if ds.name:
                r.write("<div class='dataset-name'>%s</div>\n\n" % ds.name)

            if ds.tags or ds.tasks:
                r.write("<div class='tagtask'>")
                if ds.tags:
                    r.write(
                        "".join("<span class='tag'>%s</span>" % tag for tag in ds.tags)
                    )
                if ds.tasks:
                    r.write(
                        "".join(
                            "<span class='task'>%s</span>" % task for task in ds.tasks
                        )
                    )
                r.write("</div>")

            if ds.url:
                r.write("""<div><a href="{0}">{0}</a></div>""".format(ds.url))
            if ds.description:
                r.write("<div class='description'>%s</div>" % ds.description)

            if self.experimaestro_plugin and ds.configtype:
                qualname, href = self.experimaestro_plugin.getConfigLink(
                    page.url, ds.configtype
                )
                if href:
                    r.write(
                        f"""<div><b>Type</b>: <a href="{href}">{qualname}</a></div>"""
                    )
                else:
                    r.write(f"<div><b>Type</b>: {qualname}</div>")

            r.write("</div>")

        return r.getvalue()



# flake8: noqa: F401 (re-exports)
from .context import (
    Context,
    Repository,
    BaseRepository,
    get_dataset,
    prepare_dataset,
)

from .definitions import dataset, metadata
from .data import Base
from .version import __version__



"""Record module for type-safe heterogeneous containers.

.. deprecated:: 2.0
    This module will be removed in v2. Use :class:`typing.TypedDict` instead
    for type-safe heterogeneous data structures. TypedDict provides better IDE
    support, type checking, and is part of the standard library.

    When using TypedDict, define key constants in classes (e.g., ``MyItem.ID``)
    to avoid typos and enable IDE autocomplete. Prefix keys with package name
    using underscore ``_`` as delimiter to avoid conflicts between different
    data sources.

Example migration::

    # Old way (deprecated)
    @define
    class MyItem(Item):
        value: int

    record = Record(MyItem(42))
    print(record[MyItem].value)

    # New way (recommended)
    from typing import TypedDict

    # Define key constants in classes
    class MyItem:
        ID = "mypackage_value"

    class MyRecord(TypedDict):
        mypackage_value: int

    data: MyRecord = {MyItem.ID: 42}
    print(data[MyItem.ID])
"""

import warnings
from typing import Type, TypeVar, Dict, Union, Optional

# Emit deprecation warning when module is imported
warnings.warn(
    "The datamaestro.record module is deprecated and will be removed in v2. "
    "Use typing.TypedDict instead (use class constants like MyItem.ID for keys, "
    "prefixed with package name).",
    DeprecationWarning,
    stacklevel=2,
)


class Item:
    """Base class for all item types"""

    @classmethod
    def __get_base__(cls: Type) -> Type:
        """Get the most generic superclass for this type of item"""
        if base := getattr(cls, "__base__cache__", None):
            return base

        base = cls
        for supercls in cls.__mro__:
            if issubclass(supercls, Item) and supercls is not Item:
                base = supercls
        setattr(cls, "__base__cache__", base)
        return base


T = TypeVar("T", bound=Item)
Items = Dict[Type[T], T]


class RecordType:
    def __init__(self, *item_types: Type[T]):
        self.item_types = frozenset(item_types)
        self.mapping = {item_type.__get_base__(): item_type for item_type in item_types}

    def __repr__(self):
        names = ",".join(item_type.__name__ for item_type in self.item_types)
        return f"Record({names})"

    def contains(self, other: "RecordType"):
        """Checks that each item type in other has an item type of a compatible
        type in self"""
        if len(self.item_types) != len(other.item_types):
            return False

        for item_type in other.item_types:
            if matching_type := self.mapping.get(item_type.__get_base__(), None):
                if not issubclass(matching_type, item_type):
                    return False
            else:
                return False

        return True

    def sub(self, *item_types: Type[T]):
        """Returns a new record type based on self and new item types"""
        cls_itemtypes = [x for x in self.item_types]
        mapping = {
            itemtype.__get_base__(): ix for ix, itemtype in enumerate(cls_itemtypes)
        }

        for itemtype in item_types:
            if (ix := mapping.get(itemtype.__get_base__(), -1)) >= 0:
                cls_itemtypes[ix] = itemtype
            else:
                cls_itemtypes.append(itemtype)

        return record_type(*cls_itemtypes)

    def __call__(self, *items: T):
        record = Record(*items)
        self.validate(record)
        return record

    def has(self, itemtype: Type[T]):
        return issubclass(self.mapping[itemtype.__get_base__()], itemtype)

    def validate(self, record: "Record"):
        """Creates and validate a new record of this type"""
        if self.item_types:
            for item_type in self.item_types:
                try:
                    record.__getitem__(item_type)
                except KeyError:
                    raise KeyError(f"Item of type {item_type} is missing")

        if len(record.items) != len(self.item_types):
            unregistered = [
                item
                for item in record.items.values()
                if all(
                    not issubclass(item.__get_base__(), item_type)
                    for item_type in self.item_types
                )
            ]
            raise KeyError(
                f"The record of type {self} contains unregistered items: {unregistered}"
            )

        # Creates a new record
        return record


def record_type(*item_types: Type[T]):
    """Returns a new record type"""
    return RecordType(*item_types)


class Record:
    """Associate types with entries

    A record is a composition of items; each item base class is unique.
    """

    #: Items for this record
    items: Items

    def __init__(self, *items: Union[Items, T], override=False):
        self.items = {}

        if len(items) == 1 and isinstance(items[0], dict):
            # Just copy the dictionary
            self.items = items[0]
        else:
            for entry in items:
                # Returns a new record if the item exists
                base = entry.__get_base__()
                if not override and base in self.items:
                    raise RuntimeError(
                        f"The item type {base} ({entry.__class__})"
                        " is already in the record"
                    )
                self.items[base] = entry

    def __str__(self):
        return (
            "{"
            + ", ".join(
                f"{key.__module__}.{key.__qualname__}: {value}"
                for key, value in self.items.items()
            )
            + "}"
        )

    def __repr__(self):
        return (
            "{"
            + ", ".join(
                f"{key.__module__}.{key.__qualname__}: {repr(value)}"
                for key, value in self.items.items()
            )
            + "}"
        )

    def get(self, key: Type[T]) -> Optional[T]:
        """Get a given item or None if it does not exist"""
        try:
            return self[key]
        except KeyError:
            return None

    def has(self, key: Type[T]) -> bool:
        """Returns True if the record has the given item type"""
        return key.__get_base__() in self.items

    def __getitem__(self, key: Type[T]) -> T:
        """Get an item given its type"""
        base = key.__get_base__()
        try:
            entry = self.items[base]
        except KeyError:
            raise KeyError(
                f"""No entry with type {key}: """
                f"""{",".join(str(s) for s in self.items.keys())}"""
            )

        # Check if this matches the expected class
        if not isinstance(entry, key):
            raise KeyError(
                f"""No entry with type {key}: """
                f"""{",".join(str(s) for s in self.items.keys())}"""
            )
        return entry

    def update(self, *items: T, target: RecordType = None) -> "Record":
        """Update some items"""
        # Create our new dictionary
        item_dict = {**self.items}
        for item in items:
            item_dict[item.__get_base__()] = item

        return Record(item_dict)



"""Machine learning generic data formats"""

from pathlib import Path
from typing import Generic, TypeVar, Optional
from experimaestro import Param, Meta
from . import Base

Train = TypeVar("Train", bound=Base)
Validation = TypeVar("Validation", bound=Base)
Test = TypeVar("Test", bound=Base)


class Supervised(Base, Generic[Train, Validation, Test]):
    train: Param[Base]
    """The training dataset"""

    validation: Param[Optional[Base]] = None
    """The validation dataset (optional)"""

    test: Param[Optional[Base]] = None
    """The training optional"""


class FolderBased(Base):
    """Classification dataset where folders give the basis"""

    classes: Param[list[str]]
    path: Meta[Path]



"""HuggingFace Hub download resources.

Provides a ValueResource subclass for loading datasets from
the HuggingFace Hub.
"""

from __future__ import annotations

import logging

from datamaestro.download import ValueResource

logger = logging.getLogger(__name__)


class HFDownloader(ValueResource):
    """Load a dataset from the HuggingFace Hub.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = HFDownloader.apply(
                "hf_data", repo_id="user/dataset"
            )

    Usage as decorator (deprecated)::

        @hf_download("hf_data", repo_id="user/dataset")
        @dataset(Base)
        def my_dataset(hf_data): ...
    """

    def __init__(
        self,
        varname: str,
        repo_id: str,
        *,
        data_files: str | None = None,
        split: str | None = None,
        transient: bool = False,
    ):
        """
        Args:
            varname: Variable name.
            repo_id: The HuggingFace repository ID.
            data_files: Specific data files to load.
            split: Dataset split to load.
            transient: If True, data can be deleted after dependents
                complete.
        """
        super().__init__(varname=varname, transient=transient)
        self.repo_id = repo_id
        self.data_files = data_files
        self.split = split

    def download(self, force=False):
        try:
            from datasets import load_dataset
        except ModuleNotFoundError:
            logger.error("the datasets library is not installed:")
            logger.error("pip install datasets")
            raise

        self._dataset = load_dataset(self.repo_id, data_files=self.data_files)
        return True

    def prepare(self):
        return {
            "repo_id": self.repo_id,
            "data_files": self.data_files,
            "split": self.split,
        }


# Factory alias for backward compat
hf_download = HFDownloader.apply



"""Tests for the new Resource interface.

Covers:
- ResourceState enum and metadata persistence
- Resource base class (bind, dependencies, state, cleanup)
- FileResource, FolderResource, ValueResource
- Topological sort and cycle detection
- Two-path download flow (transient_path -> path)
- Eager transient cleanup
- can_recover property behavior
- Both new class-based and legacy decorator-based dataset definitions
- Each concrete resource type
"""

from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from datamaestro.definitions import (
    AbstractDataset,
    topological_sort,
    _compute_dependents,
    _bind_class_resources,
)
from datamaestro.download import (
    Resource,
    ResourceState,
    ResourceStateFile,
    FileResource,
    FolderResource,
    ValueResource,
    Download,
    reference,
)
from .conftest import MyRepository


# ---- Helpers ----


class SimpleDataset(AbstractDataset):
    """Minimal dataset for testing."""

    def __init__(self, repository, datapath: Path):
        super().__init__(repository)
        self._datapath = datapath

    @property
    def datapath(self):
        return self._datapath

    def _prepare(self):
        # Return a mock Base-like object for the prepare flow
        obj = MagicMock()
        obj.__xpm__ = MagicMock()
        obj.__xpm__.values = {}
        return obj

    @property
    def description(self):
        return "test dataset"


class DummyFileResource(FileResource):
    """Concrete FileResource for testing."""

    def __init__(self, filename, url="http://example.com/test", **kw):
        super().__init__(filename, **kw)
        self.url = url
        self._download_called = False

    def _download(self, destination: Path):
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text(f"downloaded from {self.url}")
        self._download_called = True


class DummyFolderResource(FolderResource):
    """Concrete FolderResource for testing."""

    def __init__(self, **kw):
        super().__init__(**kw)
        self._download_called = False

    def _download(self, destination: Path):
        destination.mkdir(parents=True, exist_ok=True)
        (destination / "file.txt").write_text("content")
        self._download_called = True


class DummyValueResource(ValueResource):
    """Concrete ValueResource for testing."""

    def __init__(self, value, **kw):
        super().__init__(**kw)
        self._value = value

    def download(self, force=False):
        pass

    def prepare(self):
        return self._value


class RecoverableResource(FileResource):
    """Resource that supports recovery from PARTIAL state."""

    @property
    def can_recover(self) -> bool:
        return True

    def __init__(self, filename, **kw):
        super().__init__(filename, **kw)

    def _download(self, destination: Path):
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text("recovered")


class FailingResource(FileResource):
    """Resource that fails during download."""

    def __init__(self, filename, **kw):
        super().__init__(filename, **kw)

    def _download(self, destination: Path):
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text("partial data")
        raise RuntimeError("Download failed")


class DependentResource(FileResource):
    """Resource that depends on another resource."""

    def __init__(self, filename, source: Resource, **kw):
        super().__init__(filename, **kw)
        self._dependencies = [source]

    def _download(self, destination: Path):
        # Read from dependency's path
        source = self.dependencies[0]
        data = source.path.read_text()
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text(f"processed: {data}")


# ---- Fixtures ----


@pytest.fixture
def datapath(tmp_path):
    """Temporary dataset data path."""
    return tmp_path / "dataset"


@pytest.fixture
def dataset(context, datapath):
    """A minimal dataset bound to a repository."""
    repository = MyRepository(context)
    ds = SimpleDataset(repository, datapath)
    return ds


# ==== ResourceState Tests ====


class TestResourceState:
    def test_values(self):
        assert ResourceState.NONE == "none"
        assert ResourceState.PARTIAL == "partial"
        assert ResourceState.COMPLETE == "complete"

    def test_from_string(self):
        assert ResourceState("none") == ResourceState.NONE
        assert ResourceState("partial") == ResourceState.PARTIAL
        assert ResourceState("complete") == ResourceState.COMPLETE


# ==== ResourceStateFile Tests ====


class TestResourceStateFile:
    def test_read_nonexistent(self, datapath):
        sf = ResourceStateFile(datapath)
        assert sf.read("TRAIN") == ResourceState.NONE

    def test_write_and_read(self, datapath):
        sf = ResourceStateFile(datapath)
        sf.write("TRAIN", ResourceState.COMPLETE)

        assert sf.read("TRAIN") == ResourceState.COMPLETE
        assert sf.read("TEST") == ResourceState.NONE

    def test_multiple_resources(self, datapath):
        sf = ResourceStateFile(datapath)
        sf.write("A", ResourceState.COMPLETE)
        sf.write("B", ResourceState.PARTIAL)
        sf.write("C", ResourceState.NONE)

        assert sf.read("A") == ResourceState.COMPLETE
        assert sf.read("B") == ResourceState.PARTIAL
        assert sf.read("C") == ResourceState.NONE

    def test_overwrite(self, datapath):
        sf = ResourceStateFile(datapath)
        sf.write("A", ResourceState.PARTIAL)
        assert sf.read("A") == ResourceState.PARTIAL

        sf.write("A", ResourceState.COMPLETE)
        assert sf.read("A") == ResourceState.COMPLETE

    def test_file_format(self, datapath):
        sf = ResourceStateFile(datapath)
        sf.write("TRAIN", ResourceState.COMPLETE)

        state_path = datapath / ".state.json"
        assert state_path.exists()

        with state_path.open() as f:
            data = json.load(f)

        assert data["version"] == 1
        assert data["resources"]["TRAIN"]["state"] == "complete"


# ==== Resource Base Class Tests ====


class TestResourceBase:
    def test_bind(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        assert r.name == "TEST"
        assert r.dataset is dataset
        assert "TEST" in dataset.resources
        assert r in dataset.ordered_resources

    def test_bind_with_varname(self, dataset):
        r = DummyFileResource("test.txt", varname="my_var")
        r.bind("ATTR_NAME", dataset)

        # varname takes precedence
        assert r.name == "my_var"

    def test_bind_duplicate_raises(self, dataset):
        r1 = DummyFileResource("test1.txt")
        r2 = DummyFileResource("test2.txt")
        r1.bind("TEST", dataset)

        with pytest.raises(AssertionError, match="already declared"):
            r2.bind("TEST", dataset)

    def test_bind_already_bound_raises(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        ds2 = SimpleDataset(None, dataset.datapath / "other")
        with pytest.raises(AssertionError, match="already bound"):
            r.bind("TEST2", ds2)

    def test_state_default_none(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.state == ResourceState.NONE

    def test_state_set_and_get(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        r.state = ResourceState.COMPLETE
        assert r.state == ResourceState.COMPLETE

        r.state = ResourceState.PARTIAL
        assert r.state == ResourceState.PARTIAL

    def test_dependencies_default_empty(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.dependencies == []

    def test_dependents_default_empty(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.dependents == []

    def test_can_recover_default_false(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.can_recover is False

    def test_can_recover_override(self, dataset):
        r = RecoverableResource("test.txt")
        r.bind("TEST", dataset)
        assert r.can_recover is True

    def test_has_files_default_true(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.has_files() is True

    def test_transient_flag(self, dataset):
        r = DummyFileResource("test.txt", transient=True)
        r.bind("TEST", dataset)
        assert r.transient is True

    def test_context_property(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)
        assert r.context is dataset.context

    def test_cleanup(self, dataset):
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        # Create files at both paths
        r.path.parent.mkdir(parents=True, exist_ok=True)
        r.path.write_text("final")
        r.transient_path.parent.mkdir(parents=True, exist_ok=True)
        r.transient_path.write_text("temp")
        r.state = ResourceState.COMPLETE

        r.cleanup()

        assert not r.path.exists()
        assert not r.transient_path.exists()
        assert r.state == ResourceState.NONE


# ==== FileResource Tests ====


class TestFileResource:
    def test_path(self, dataset):
        r = DummyFileResource("data.csv")
        r.bind("DATA", dataset)

        expected = dataset.datapath / "data.csv"
        assert r.path == expected

    def test_transient_path(self, dataset):
        r = DummyFileResource("data.csv")
        r.bind("DATA", dataset)

        expected = dataset.datapath / ".downloads" / "data.csv"
        assert r.transient_path == expected

    def test_prepare_returns_path(self, dataset):
        r = DummyFileResource("data.csv")
        r.bind("DATA", dataset)
        assert r.prepare() == r.path

    def test_download_writes_to_transient(self, dataset):
        r = DummyFileResource("data.csv")
        r.bind("DATA", dataset)
        r.download()

        assert r.transient_path.exists()
        assert "downloaded" in r.transient_path.read_text()
        assert r._download_called

    def test_stream_default_none(self, dataset):
        r = DummyFileResource("data.csv")
        r.bind("DATA", dataset)
        assert r.stream() is None

    def test_varname_from_filename(self):
        """Without explicit varname, name is derived from filename."""
        r = DummyFileResource("data.csv.gz")
        assert r.name == "data"


# ==== FolderResource Tests ====


class TestFolderResource:
    def test_path(self, dataset):
        r = DummyFolderResource(varname="archive")
        r.bind("ARCHIVE", dataset)

        expected = dataset.datapath / "archive"
        assert r.path == expected

    def test_transient_path(self, dataset):
        r = DummyFolderResource(varname="archive")
        r.bind("ARCHIVE", dataset)

        expected = dataset.datapath / ".downloads" / "archive"
        assert r.transient_path == expected

    def test_prepare_returns_path(self, dataset):
        r = DummyFolderResource(varname="archive")
        r.bind("ARCHIVE", dataset)
        assert r.prepare() == r.path

    def test_download_creates_directory(self, dataset):
        r = DummyFolderResource(varname="archive")
        r.bind("ARCHIVE", dataset)
        r.download()

        assert r.transient_path.is_dir()
        assert (r.transient_path / "file.txt").exists()


# ==== ValueResource Tests ====


class TestValueResource:
    def test_has_files_false(self, dataset):
        r = DummyValueResource({"key": "value"}, varname="data")
        r.bind("DATA", dataset)
        assert r.has_files() is False

    def test_prepare_returns_value(self, dataset):
        val = {"key": "value"}
        r = DummyValueResource(val, varname="data")
        r.bind("DATA", dataset)
        assert r.prepare() == val


# ==== Topological Sort Tests ====


class TestTopologicalSort:
    def test_empty(self):
        assert topological_sort({}) == []

    def test_single(self, dataset):
        r = DummyFileResource("a.txt")
        r.bind("A", dataset)
        result = topological_sort(dataset.resources)
        assert result == [r]

    def test_linear_chain(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        result = topological_sort(dataset.resources)
        assert result.index(a) < result.index(b)

    def test_diamond(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        c = DependentResource("c.txt", source=a)
        c.bind("C", dataset)

        d = DependentResource("d.txt", source=b)
        d._dependencies.append(c)
        d.bind("D", dataset)

        result = topological_sort(dataset.resources)
        assert result.index(a) < result.index(b)
        assert result.index(a) < result.index(c)
        assert result.index(b) < result.index(d)
        assert result.index(c) < result.index(d)

    def test_cycle_detection(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        # Create cycle: a depends on b
        a._dependencies = [b]

        with pytest.raises(ValueError, match="Cycle detected"):
            topological_sort(dataset.resources)

    def test_independent_resources(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DummyFileResource("b.txt")
        b.bind("B", dataset)

        result = topological_sort(dataset.resources)
        assert len(result) == 2
        assert set(result) == {a, b}


# ==== Dependents Computation Tests ====


class TestComputeDependents:
    def test_no_dependencies(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        _compute_dependents(dataset.resources)
        assert a.dependents == []

    def test_linear_dependents(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        _compute_dependents(dataset.resources)
        assert b in a.dependents
        assert a not in b.dependents

    def test_multiple_dependents(self, dataset):
        a = DummyFileResource("a.txt")
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        c = DependentResource("c.txt", source=a)
        c.bind("C", dataset)

        _compute_dependents(dataset.resources)
        assert set(a.dependents) == {b, c}


# ==== Two-Path Download Flow Tests ====


class TestTwoPathFlow:
    def test_download_moves_to_final_path(self, dataset):
        """Framework should move transient_path -> path on success."""
        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        dataset.download()

        assert r.path.exists()
        assert r.state == ResourceState.COMPLETE

    def test_failure_no_recover_cleans_up(self, dataset):
        """On failure without can_recover, transient data is deleted."""
        r = FailingResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        result = dataset.download()

        assert result is False
        assert not r.transient_path.exists()
        assert r.state == ResourceState.NONE

    def test_failure_with_recover_preserves(self, dataset):
        """On failure with can_recover, transient data is preserved."""

        class FailRecoverable(RecoverableResource):
            def _download(self, destination):
                destination.parent.mkdir(parents=True, exist_ok=True)
                destination.write_text("partial")
                raise RuntimeError("partial failure")

        r = FailRecoverable("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        result = dataset.download()

        assert result is False
        assert r.transient_path.exists()
        assert r.state == ResourceState.PARTIAL

    def test_skip_complete_resources(self, dataset):
        """Resources already COMPLETE are skipped unless force=True."""
        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        # Mark as complete
        r.state = ResourceState.COMPLETE
        r.path.parent.mkdir(parents=True, exist_ok=True)
        r.path.write_text("existing")

        dataset.download()

        # download should not have been called
        assert r._download_called is False

    def test_redownload_when_files_missing(self, dataset):
        """COMPLETE resource with missing files is re-downloaded."""
        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        # Mark as complete but do NOT create the file
        r.state = ResourceState.COMPLETE
        assert not r.path.exists()

        dataset.download()

        # Should have re-downloaded
        assert r._download_called is True
        assert r.path.exists()
        assert r.state == ResourceState.COMPLETE

    def test_adopt_preexisting_files(self, dataset):
        """Files already on disk (old downloads) are adopted as COMPLETE."""
        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        # Pre-create the file at the final path (simulating old download)
        r.path.parent.mkdir(parents=True, exist_ok=True)
        r.path.write_text("old data")

        # State is NONE (no .state.json entry)
        assert r.state == ResourceState.NONE

        dataset.download()

        # Should NOT have re-downloaded — just marked COMPLETE
        assert r._download_called is False
        assert r.state == ResourceState.COMPLETE
        assert r.path.read_text() == "old data"

    def test_downloads_dir_cleaned_after_success(self, dataset):
        """The .downloads/ directory is removed after all succeed."""
        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        result = dataset.download()

        assert result is True
        downloads_dir = dataset.datapath / ".downloads"
        assert not downloads_dir.exists()

    def test_downloads_dir_kept_on_failure(self, dataset):
        """The .downloads/ directory is kept if a download fails."""
        r = FailingResource("data.txt")
        r.bind("DATA", dataset)

        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        # Pre-create .downloads/ with transient data
        r.transient_path.parent.mkdir(parents=True, exist_ok=True)
        r.transient_path.write_text("partial")

        result = dataset.download()

        assert result is False
        # .downloads/ should still exist (failure, no cleanup)
        # (transient data itself is deleted because can_recover=False)

    def test_lock_prevents_concurrent_download(self, dataset):
        """A second download blocks while the first holds the lock."""
        import fcntl
        import threading

        r = DummyFileResource("data.txt")
        r.bind("DATA", dataset)
        dataset.ordered_resources = [r]
        _compute_dependents(dataset.resources)

        # Acquire the lock externally to simulate a concurrent download
        dataset.datapath.mkdir(parents=True, exist_ok=True)
        lock_path = dataset.datapath / ".state.lock"
        lock_file = lock_path.open("w")
        fcntl.flock(lock_file, fcntl.LOCK_EX)

        result_holder = {}

        def try_download():
            result_holder["result"] = dataset.download()

        t = threading.Thread(target=try_download)
        t.start()
        # Give thread time to hit the lock
        t.join(timeout=0.2)
        # Thread should still be alive (blocked on lock)
        assert t.is_alive()

        # Release the lock
        fcntl.flock(lock_file, fcntl.LOCK_UN)
        lock_file.close()

        t.join(timeout=5)
        assert not t.is_alive()
        assert result_holder["result"] is True


# ==== Eager Transient Cleanup Tests ====


class TestTransientCleanup:
    def test_transient_cleaned_after_dependents_complete(self, dataset):
        """Transient resources are cleaned up when all dependents
        are COMPLETE."""
        a = DummyFileResource("a.txt", transient=True)
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        _compute_dependents(dataset.resources)
        dataset.ordered_resources = topological_sort(dataset.resources)

        dataset.download()

        # b should be complete
        assert b.state == ResourceState.COMPLETE
        assert b.path.exists()

        # a should be cleaned up (transient)
        assert a.state == ResourceState.NONE
        assert not a.path.exists()

    def test_non_transient_not_cleaned(self, dataset):
        """Non-transient resources are NOT cleaned up."""
        a = DummyFileResource("a.txt", transient=False)
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        _compute_dependents(dataset.resources)
        dataset.ordered_resources = topological_sort(dataset.resources)

        dataset.download()

        assert a.state == ResourceState.COMPLETE
        assert a.path.exists()

    def test_transient_not_cleaned_if_dependent_incomplete(self, dataset):
        """Transient resources are NOT cleaned if a dependent
        hasn't completed yet."""
        a = DummyFileResource("a.txt", transient=True)
        a.bind("A", dataset)

        b = DependentResource("b.txt", source=a)
        b.bind("B", dataset)

        c = DependentResource("c.txt", source=a)
        c.bind("C", dataset)

        _compute_dependents(dataset.resources)
        dataset.ordered_resources = topological_sort(dataset.resources)

        # Download only processes in order, so after B completes,
        # C hasn't yet — a should not be cleaned up until C completes.
        # The full download() handles this correctly.
        dataset.download()

        # After full download, all dependents are complete
        # so transient should be cleaned
        assert a.state == ResourceState.NONE


# ==== Legacy Decorator-Based Dataset Tests ====


class TestLegacyDecoratorDataset:
    def test_filedownloader_decorator(self, context):
        """Legacy decorator-based filedownloader still works."""
        import warnings
        from datamaestro.download.single import filedownloader

        repository = MyRepository(context)
        ds = SimpleDataset(repository, context.datapath / "legacy")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            downloader = filedownloader("test.html", "http://httpbin.org/html")
            downloader(ds)

        assert "test" in ds.resources
        assert ds.resources["test"] is downloader

    def test_reference_resource(self, context):
        """reference resource still works."""
        repository = MyRepository(context)
        ds = SimpleDataset(repository, context.datapath / "ref_test")

        mock_ref = MagicMock()
        mock_ref.prepare.return_value = "prepared_value"

        ref = reference(varname="ref", reference=mock_ref)
        ref.bind("ref", ds)

        assert ref.has_files() is False
        result = ref.prepare()
        assert result == "prepared_value"


# ==== New Class-Based Dataset Tests ====


class TestClassBasedDataset:
    def test_bind_class_resources(self, dataset):
        """_bind_class_resources detects Resource attributes."""
        from datamaestro.data import Base

        class MyData(Base):
            A = DummyFileResource("a.txt")
            B = DummyFileResource("b.txt")

        _bind_class_resources(MyData, dataset)

        assert "A" in dataset.resources
        assert "B" in dataset.resources
        assert len(dataset.ordered_resources) == 2

    def test_bind_with_dependencies(self, dataset):
        """Resources with dependencies are properly ordered."""
        from datamaestro.data import Base

        src = DummyFileResource("src.txt")

        class MyData(Base):
            SRC = src
            PROCESSED = DependentResource("proc.txt", source=src)

        _bind_class_resources(MyData, dataset)

        # SRC should come before PROCESSED in ordered_resources
        src_idx = dataset.ordered_resources.index(MyData.SRC)
        proc_idx = dataset.ordered_resources.index(MyData.PROCESSED)
        assert src_idx < proc_idx

        # Check dependents were computed
        assert MyData.PROCESSED in MyData.SRC.dependents

    def test_non_resource_attributes_ignored(self, dataset):
        """Non-Resource class attributes are not bound."""
        from datamaestro.data import Base

        class MyData(Base):
            A = DummyFileResource("a.txt")
            NOT_A_RESOURCE = "just a string"
            ALSO_NOT = 42

        _bind_class_resources(MyData, dataset)

        assert "A" in dataset.resources
        assert "NOT_A_RESOURCE" not in dataset.resources
        assert "ALSO_NOT" not in dataset.resources


# ==== Backward Compatibility Tests ====


class TestBackwardCompat:
    def test_hasfiles_deprecated(self, dataset):
        """hasfiles() still works but emits deprecation."""
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        import warnings

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = r.hasfiles()

        assert result is True

    def test_definition_property_deprecated(self, dataset):
        """definition property still works but emits deprecation."""
        r = DummyFileResource("test.txt")
        r.bind("TEST", dataset)

        import warnings

        # Clear the one-time warning cache
        from datamaestro.download import _deprecation_warned

        _deprecation_warned.discard("definition")

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = r.definition

        assert result is dataset

    def test_download_subclass_deprecated(self):
        """Subclassing Download emits deprecation."""
        from datamaestro.download import _deprecation_warned

        _deprecation_warned.discard("Download-TestSub")

        import warnings

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")

            class TestSub(Download):
                def download(self, force=False):
                    pass

                def prepare(self):
                    pass

    def test_apply_classmethod(self):
        """Resource.apply creates instances."""
        r = DummyFileResource.apply("test.txt")
        assert isinstance(r, DummyFileResource)
        assert r.filename == "test.txt"


# ==== Concrete Resource Tests ====


class TestFileDownloader:
    def test_construction(self):
        """FileDownloader can be constructed."""
        from datamaestro.download.single import FileDownloader

        r = FileDownloader("data.csv", "http://example.com/data.csv")
        assert r.filename == "data.csv"
        assert r.url == "http://example.com/data.csv"
        assert r.name == "data"  # derived from filename

    def test_factory_alias(self):
        """filedownloader is an alias for FileDownloader.apply."""
        from datamaestro.download.single import (
            filedownloader,
            FileDownloader,
        )

        r = filedownloader("data.csv", "http://example.com/data.csv")
        assert isinstance(r, FileDownloader)

    def test_transient_flag(self):
        """FileDownloader accepts transient flag."""
        from datamaestro.download.single import FileDownloader

        r = FileDownloader(
            "data.csv",
            "http://example.com/data.csv",
            transient=True,
        )
        assert r.transient is True

    def test_backward_compat_alias(self):
        """SingleDownload is an alias for FileDownloader."""
        from datamaestro.download.single import (
            SingleDownload,
            FileDownloader,
        )

        assert SingleDownload is FileDownloader


class TestConcatDownloader:
    def test_construction(self):
        from datamaestro.download.single import ConcatDownloader

        r = ConcatDownloader("data.txt", "http://example.com/data.tar.gz")
        assert r.filename == "data.txt"
        assert r.url == "http://example.com/data.tar.gz"

    def test_factory_alias(self):
        from datamaestro.download.single import (
            concatdownload,
            ConcatDownloader,
        )

        r = concatdownload("data.txt", "http://example.com/data.tar.gz")
        assert isinstance(r, ConcatDownloader)


class TestArchiveDownloaders:
    def test_zip_construction(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader("archive", "http://example.com/data.zip")
        assert r.url == "http://example.com/data.zip"
        assert r.name == "archive"

    def test_tar_construction(self):
        from datamaestro.download.archive import TarDownloader

        r = TarDownloader("archive", "http://example.com/data.tar.gz")
        assert r.url == "http://example.com/data.tar.gz"

    def test_zip_factory_alias(self):
        from datamaestro.download.archive import (
            zipdownloader,
            ZipDownloader,
        )

        r = zipdownloader("archive", "http://example.com/data.zip")
        assert isinstance(r, ZipDownloader)

    def test_tar_factory_alias(self):
        from datamaestro.download.archive import (
            tardownloader,
            TarDownloader,
        )

        r = tardownloader("archive", "http://example.com/data.tar.gz")
        assert isinstance(r, TarDownloader)


class TestCustomDownload:
    def test_construction(self):
        from datamaestro.download.custom import custom_download

        fn = MagicMock()
        r = custom_download("data", fn)
        assert r.name == "data"
        assert r.downloader is fn


class TestHFDownloader:
    def test_construction(self):
        from datamaestro.download.huggingface import HFDownloader

        r = HFDownloader("hf", repo_id="user/dataset")
        assert r.repo_id == "user/dataset"
        assert r.name == "hf"

    def test_factory_alias(self):
        from datamaestro.download.huggingface import (
            hf_download,
            HFDownloader,
        )

        r = hf_download("hf", repo_id="user/dataset")
        assert isinstance(r, HFDownloader)

    def test_prepare(self):
        from datamaestro.download.huggingface import HFDownloader

        r = HFDownloader(
            "hf",
            repo_id="user/dataset",
            data_files="train.csv",
            split="train",
        )
        result = r.prepare()
        assert result == {
            "repo_id": "user/dataset",
            "data_files": "train.csv",
            "split": "train",
        }


class TestTodoResource:
    def test_raises_not_implemented(self):
        from datamaestro.download.todo import Todo

        r = Todo(varname="test")
        with pytest.raises(NotImplementedError):
            r.download()

        with pytest.raises(NotImplementedError):
            r.prepare()


class TestReferenceResource:
    def test_has_files_false(self, dataset):
        mock_ref = MagicMock()
        mock_ref.prepare.return_value = "value"

        r = reference(varname="ref", reference=mock_ref)
        r.bind("ref", dataset)

        assert r.has_files() is False

    def test_prepare_delegates(self, dataset):
        mock_ref = MagicMock()
        mock_ref.prepare.return_value = "prepared"

        r = reference(varname="ref", reference=mock_ref)
        r.bind("ref", dataset)

        result = r.prepare()
        assert result == "prepared"

    def test_download_delegates(self, dataset):
        mock_ref = MagicMock()
        mock_ref.__datamaestro__ = MagicMock()

        r = reference(varname="ref", reference=mock_ref)
        r.bind("ref", dataset)

        r.download(force=True)
        mock_ref.__datamaestro__.download.assert_called_once_with(True)

    def test_requires_reference(self):
        with pytest.raises(AssertionError, match="cannot be null"):
            reference(varname="ref", reference=None)


# ==== Links Resource Tests ====


class TestLinksResource:
    def test_construction(self):
        from datamaestro.download.links import links

        mock_ds = MagicMock()
        r = links("data", ref1=mock_ds)
        assert r.name == "data"

    def test_has_files_false(self, dataset):
        from datamaestro.download.links import links

        mock_ds = MagicMock()
        r = links("data", ref1=mock_ds)
        r.bind("data", dataset)

        assert r.has_files() is False

    def test_path_is_datapath(self, dataset):
        from datamaestro.download.links import links

        mock_ds = MagicMock()
        r = links("data", ref1=mock_ds)
        r.bind("data", dataset)

        assert r.path == dataset.datapath

    def test_prepare_returns_path(self, dataset):
        from datamaestro.download.links import links

        mock_ds = MagicMock()
        r = links("data", ref1=mock_ds)
        r.bind("data", dataset)

        assert r.prepare() == dataset.datapath


class TestLinkFolder:
    def test_construction(self):
        from datamaestro.download.links import linkfolder

        r = linkfolder("data", proposals=["/tmp/test"])
        assert r.name == "data"

    def test_check_is_dir(self, dataset, tmp_path):
        from datamaestro.download.links import linkfolder

        r = linkfolder("data", proposals=[])
        r.bind("data", dataset)

        # A directory should pass
        assert r.check(tmp_path) is True
        # A non-existent path should fail
        assert r.check(tmp_path / "nonexistent") is False

    def test_path(self, dataset):
        from datamaestro.download.links import linkfolder

        r = linkfolder("data", proposals=[])
        r.bind("data", dataset)

        assert r.path == dataset.datapath / "data"

    def test_prepare_returns_path(self, dataset):
        from datamaestro.download.links import linkfolder

        r = linkfolder("data", proposals=[])
        r.bind("data", dataset)

        assert r.prepare() == r.path


class TestLinkFile:
    def test_construction(self):
        from datamaestro.download.links import linkfile

        r = linkfile("data", proposals=["/tmp/test.txt"])
        assert r.name == "data"

    def test_check_is_file(self, dataset, tmp_path):
        from datamaestro.download.links import linkfile

        r = linkfile("data", proposals=[])
        r.bind("data", dataset)

        # Create a real file to check
        test_file = tmp_path / "test.txt"
        test_file.write_text("hello")

        assert r.check(test_file) is True
        # A directory should fail
        assert r.check(tmp_path) is False
        # A non-existent path should fail
        assert r.check(tmp_path / "nonexistent") is False

    def test_path(self, dataset):
        from datamaestro.download.links import linkfile

        r = linkfile("data", proposals=[])
        r.bind("data", dataset)

        assert r.path == dataset.datapath / "data"


# ==== Wayback Resource Tests ====


class TestWaybackDocuments:
    def test_construction(self):
        from datamaestro.download.wayback import wayback_documents

        def urls_fn():
            return iter(["http://example.com"])

        r = wayback_documents("20200101", urls_fn, name="wb")
        assert r.name == "wb"
        assert r.timestamp == "20200101"

    def test_prepare_returns_path(self, dataset):
        from datamaestro.download.wayback import wayback_documents

        def urls_fn():
            return iter([])

        r = wayback_documents("20200101", urls_fn, name="wb")
        r.bind("wb", dataset)

        expected = dataset.datapath / "wb"
        assert r.prepare() == expected


# ==== Custom Download Functional Tests ====


class TestCustomDownloadFunctional:
    def test_download_delegates(self, dataset):
        from datamaestro.download.custom import custom_download

        fn = MagicMock()
        r = custom_download("data", fn)
        r.bind("data", dataset)

        r.download(force=True)

        fn.assert_called_once_with(dataset.context, dataset.datapath, force=True)

    def test_prepare_returns_datapath(self, dataset):
        from datamaestro.download.custom import custom_download

        fn = MagicMock()
        r = custom_download("data", fn)
        r.bind("data", dataset)

        assert r.prepare() == dataset.datapath


# ==== Archive Downloader Base Tests ====


class TestArchiveDownloaderBase:
    def test_zip_path_with_postinit(self, dataset):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader("archive", "http://example.com/data.zip")
        r.bind("archive", dataset)

        # path should trigger postinit
        p = r.path
        assert isinstance(p, Path)

    def test_tar_path_with_postinit(self, dataset):
        from datamaestro.download.archive import TarDownloader

        r = TarDownloader("archive", "http://example.com/data.tar.gz")
        r.bind("archive", dataset)

        p = r.path
        assert isinstance(p, Path)

    def test_extractall_default(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader("archive", "http://example.com/data.zip")
        assert r.extractall is True

    def test_extractall_with_subpath(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader(
            "archive",
            "http://example.com/data.zip",
            subpath="subdir",
        )
        assert r.extractall is False

    def test_extractall_with_files(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader(
            "archive",
            "http://example.com/data.zip",
            files={"file1.txt"},
        )
        assert r.extractall is False

    def test_subpath_trailing_slash(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader(
            "archive",
            "http://example.com/data.zip",
            subpath="subdir",
        )
        assert r.subpath == "subdir/"

    def test_transient_flag(self):
        from datamaestro.download.archive import ZipDownloader

        r = ZipDownloader(
            "archive",
            "http://example.com/data.zip",
            transient=True,
        )
        assert r.transient is True


# ==== gsync (legacy) Tests ====


class TestGsync:
    def test_import(self):
        """gsync can be imported (legacy Download subclass)."""
        from datamaestro.download.sync import gsync

        assert issubclass(gsync, Download)


# ==== manual.py (deprecated re-export) Tests ====


class TestManual:
    def test_import_linkfolder(self):
        """manual.linkfolder is a deprecated re-export."""
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            from datamaestro.download.manual import linkfolder

            assert linkfolder is not None


# ==== multiple.py (legacy) Tests ====


class TestMultiple:
    def test_import_list(self):
        """List can be imported (legacy Download subclass)."""
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            from datamaestro.download.multiple import List

            assert issubclass(List, Download)

    def test_import_datasets(self):
        """Datasets can be imported (legacy Download subclass)."""
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            from datamaestro.download.multiple import Datasets

            assert issubclass(Datasets, Download)



"""Huggingface datamaestro adapters"""

from functools import cached_property
from typing import Optional
from . import Base
import logging
from experimaestro import Param


class HuggingFaceDataset(Base):
    repo_id: Param[str]
    data_files: Param[Optional[str]] = None
    split: Param[Optional[str]] = None

    @cached_property
    def data(self):
        try:
            from datasets import load_dataset
        except ModuleNotFoundError:
            logging.error("the datasets library is not installed:")
            logging.error("pip install datasets")
            raise

        ds = load_dataset(self.repo_id, data_files=self.data_files, split=self.split)
        return ds



from pathlib import Path
from datamaestro import Repository, Context
import shutil
import logging
import pytest


class MyRepository(Repository):
    NAMESPACE = "test-simple"
    AUTHOR = """Benjamin Piwowarski <benjamin@piwowarski.fr>"""
    DESCRIPTION = """Repository with tests"""


repository = None


@pytest.fixture(scope="session")
def context(tmp_path_factory):
    """Sets a temporary main directory"""
    logging.warning("In context fixture")
    dir = tmp_path_factory.mktemp("datamaestro-data")
    context = Context(Path(dir))
    logging.info("Created datamaestro test directory %s", dir)

    _repository = MyRepository(context)  # noqa: F841 - registered on creation

    yield context

    logging.info("Removing datamaestro test directory %s", dir)
    shutil.rmtree(dir)



"""Custom download resources.

Provides a Resource subclass that delegates to a user-defined
download function.
"""

from __future__ import annotations

from pathlib import Path
from typing import Protocol

from datamaestro import Context
from datamaestro.download import Resource


class Downloader(Protocol):
    def __call__(self, context: Context, root: Path, *, force: bool = False):
        pass


class custom_download(Resource):
    """A resource that delegates to a user-defined download function.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = custom_download(
                "data", downloader=my_download_fn
            )

    Usage as decorator (deprecated)::

        @custom_download("data", downloader=my_download_fn)
        @dataset(Base)
        def my_dataset(data): ...
    """

    def __init__(
        self,
        varname: str,
        downloader: Downloader,
        *,
        transient: bool = False,
    ):
        super().__init__(varname=varname, transient=transient)
        self.downloader = downloader

    def prepare(self):
        return self.dataset.datapath

    def download(self, force=False):
        self.downloader(self.context, self.dataset.datapath, force=force)



from datamaestro.annotations.agreement import useragreement
from datamaestro.definitions import AbstractDataset


def test_useragreements(context):
    # Fake dataset
    class t(AbstractDataset):
        def _prepare(self):
            pass

    useragreement("test")(t(None))



"""Link-based resources.

Provides resources that create symlinks to other datasets or
user-specified paths.
"""

from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import List

from datamaestro.context import ResolvablePath
from datamaestro.definitions import AbstractDataset
from datamaestro.download import Resource
from datamaestro.utils import deprecated

logger = logging.getLogger(__name__)


class links(Resource):
    """Link with another dataset path.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = links("data", ref1=other_dataset1)

    Usage as decorator (deprecated)::

        @links("data", ref1=other_dataset1)
        @dataset(Base)
        def my_dataset(data): ...
    """

    def __init__(
        self,
        varname: str,
        *,
        transient: bool = False,
        **link_targets: List[AbstractDataset],
    ):
        super().__init__(varname=varname, transient=transient)
        self.links = link_targets

    @property
    def path(self):
        return self.dataset.datapath

    def prepare(self):
        return self.path

    def download(self, force=False):
        self.path.mkdir(exist_ok=True, parents=True)
        for key, value in self.links.items():
            value.download(force)

            path = value()
            dest = self.path / key

            if not dest.exists():
                if dest.is_symlink():
                    logger.info("Removing dangling symlink %s", dest)
                    dest.unlink()
                os.symlink(path, dest)

    def has_files(self):
        return False


# Deprecated
Links = deprecated("Use @links instead of @Links", links)


class linkpath(Resource):
    """Link to a path selected from proposals.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = linkpath("data", proposals=[...])
    """

    def __init__(
        self,
        varname: str,
        proposals,
        *,
        transient: bool = False,
    ):
        super().__init__(varname=varname, transient=transient)
        self.proposals = proposals

    def prepare(self):
        return self.path

    @property
    def path(self):
        return self.dataset.datapath / self.name

    def download(self, force=False):
        if self.check(self.path):
            return

        if self.path.is_symlink():
            logger.warning("Removing dangling symlink %s", self.path)
            self.path.unlink()

        path = None

        for searchpath in self.proposals:
            logger.info("Trying path %s", searchpath)
            try:
                path = ResolvablePath.resolve(self.context, searchpath)
                if self.check(path):
                    break
                logger.info("Path %s not found", path)
            except KeyError:
                logger.info("Could not expand path %s", searchpath)

        # Ask the user
        while path is None or not self.check(path):
            path = Path(input("Path to %s: " % self.name))
        assert path.name

        logger.debug("Linking %s to %s", path, self.path)
        self.path.parent.mkdir(exist_ok=True, parents=True)
        os.symlink(path, self.path)

    def check(self, path):
        raise NotImplementedError()


class linkfolder(linkpath):
    """Link to a folder.

    Usage as class attribute::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = linkfolder("data", proposals=[...])
    """

    def __init__(
        self,
        varname: str,
        proposals,
        *,
        transient: bool = False,
    ):
        super().__init__(varname, proposals, transient=transient)

    def check(self, path):
        return path.is_dir()


class linkfile(linkpath):
    """Link to a file.

    Usage as class attribute::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = linkfile("data", proposals=[...])
    """

    def __init__(
        self,
        varname: str,
        proposals,
        *,
        transient: bool = False,
    ):
        super().__init__(varname, proposals, transient=transient)

    def check(self, path):
        logger.debug("Checking %s (exists: %s)", path, path.is_file())
        return path.is_file()



import logging
from pathlib import Path
from typing import Any, Dict
from experimaestro import Config, Param, Meta
from datamaestro.definitions import AbstractDataset


class Base(Config):
    """Base object for all data types"""

    id: Param[str]
    """The unique (sub-)dataset ID"""

    __datamaestro_dataset__: "AbstractDataset"

    def dataset_information(self) -> Dict[str, Any]:
        """Returns document meta-informations"""
        return {
            "id": self.id,
            "name": self.__datamaestro_dataset__.name,
            "description": self.__datamaestro_dataset__.description,
        }

    def download(self):
        """Download the dataset"""
        self.__datamaestro_dataset__.download()

    def prepare(self, *args, **kwargs):
        """Prepare the dataset"""
        self.__datamaestro_dataset__.prepare()
        return self


class Generic(Base):
    """Generic dataset

    This allows to set any value, but should only be used
    as a placeholder
    """

    def __init__(self, **kwargs):
        logging.warning("Generic should be avoided")
        super().__init__()
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)


class File(Base):
    """A data file"""

    path: Meta[Path]
    """The path of the file"""

    def open(self, mode):
        return self.path.open(mode)


class Folder(Base):
    """A data folder"""

    path: Meta[Path]

    def open(self, mode):
        return self.path.open(mode)



#
# Main datamaestro functions and data models
#

from __future__ import annotations

import logging
import inspect
import shutil
from pathlib import Path
from itertools import chain
from abc import ABC, abstractmethod
import traceback
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Callable,
    TYPE_CHECKING,
    Union,
    _GenericAlias,
)
from experimaestro import (  # noqa: F401 (re-exports)
    Param,
    Option,
    Config,
    Meta,
)
from typing import Type as TypingType  # noqa: F401 (re-exports)
from experimaestro.core.types import Type  # noqa: F401 (re-exports)

if TYPE_CHECKING:
    from .data import Base, Dataset
    from .context import Repository, Context, DatafolderPath  # noqa: F401 (re-exports)
    from datamaestro.download import Download, Resource

# --- DAG utilities ---


def topological_sort(resources: dict[str, "Resource"]) -> list["Resource"]:
    """Topological sort of resources by their dependencies.

    Args:
        resources: Dict mapping resource names to Resource instances.

    Returns:
        List of resources in dependency order (dependencies first).

    Raises:
        ValueError: If a cycle is detected in the dependency graph.
    """
    visited: set[str] = set()
    visiting: set[str] = set()  # For cycle detection
    result: list["Resource"] = []

    def visit(resource: "Resource"):
        if resource.name in visited:
            return
        if resource.name in visiting:
            raise ValueError(
                f"Cycle detected in resource dependencies involving {resource.name}"
            )

        visiting.add(resource.name)
        for dep in resource.dependencies:
            visit(dep)
        visiting.discard(resource.name)
        visited.add(resource.name)
        result.append(resource)

    for resource in resources.values():
        visit(resource)

    return result


def _compute_dependents(resources: dict[str, "Resource"]) -> None:
    """Compute the dependents (inverse edges) for all resources."""
    # Clear existing dependents
    for resource in resources.values():
        resource._dependents = []

    # Build inverse edges
    for resource in resources.values():
        for dep in resource.dependencies:
            if resource not in dep._dependents:
                dep._dependents.append(resource)


def _bind_class_resources(cls: type, dataset_wrapper: "AbstractDataset") -> None:
    """Scan class attributes for Resource instances and bind them.

    This is called when a class-based dataset is processed by the
    @dataset decorator. It detects Resource instances defined as
    class attributes and binds them to the dataset.

    Args:
        cls: The dataset class to scan.
        dataset_wrapper: The AbstractDataset to bind resources to.
    """
    from datamaestro.download import Resource

    for attr_name, attr_value in vars(cls).items():
        if isinstance(attr_value, Resource):
            attr_value.bind(attr_name, dataset_wrapper)

    # Build the dependency DAG
    _compute_dependents(dataset_wrapper.resources)

    # Validate: topological sort will raise on cycles
    dataset_wrapper.ordered_resources = topological_sort(dataset_wrapper.resources)


def _delete_path(path: Path) -> None:
    """Delete a file or directory at path."""
    if path.exists():
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()


def _move_path(src: Path, dst: Path) -> None:
    """Move a file or directory from src to dst."""
    if src.exists():
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(str(src), str(dst))


# --- Objects holding information into classes/function


class AbstractData(ABC):
    """Data definition groups common fields between a dataset and a data piece,
    such as tags and tasks"""

    def __init__(self):
        self.tags = set()
        self.tasks = set()

    def ancestors(self):
        """Returns all configuration ancestors"""
        ancestors = []
        ancestors.extend(c for c in self.base.__mro__ if issubclass(c, Config))
        return ancestors


class DataDefinition(AbstractData):
    """Object that stores the declarative part of a data(set) description"""

    def __init__(self, t, base=None):
        assert base is None or not inspect.isclass(t)

        super().__init__()

        # Copy base type and find matching repository
        self.t = t

        # Dataset id (and all aliases)
        self.id = None
        self.base = base

        self.aliases: Set[str] = set()

        self.tags = set(chain(*[c.__datamaestro__.tags for c in self.ancestors()]))
        self.tasks = set(chain(*[c.__datamaestro__.tasks for c in self.ancestors()]))
        self._description: Optional[str] = None

    @property
    def description(self):
        return self._description

    @staticmethod
    def repository_relpath(t: type) -> Tuple["Repository", List[str]]:
        """Find the repository of the current data or dataset definition"""
        from .context import Context  # noqa: F811

        repositorymap = Context.instance().repositorymap

        fullname = f"{t.__module__}.{t.__name__}"
        components = fullname.split(".")

        current: str = ""
        longest_ix = -1
        repository = None
        for ix, c in enumerate(components):
            current = f"{current}.{c}" if current else c
            if (current is not None) and (current in repositorymap):
                longest_ix = ix
                repository = repositorymap[current]

        if repository is None:
            if components[0] == "datamaestro":
                longest_ix = 0

        return repository, [s.lower() for s in components[(longest_ix + 1) :]]

    def ancestors(self):
        ancestors = []
        if self.base:
            baseclass = self.base
        else:
            baseclass = self.t

        ancestors.extend(c for c in baseclass.__mro__ if hasattr(c, "__datamaestro__"))

        return ancestors


class AbstractDataset(AbstractData):
    """Specialization of AbstractData for datasets

    A dataset:

    - has a unique ID (and aliases)
    - can be searched for
    - has a data storage space
    - has specific attributes:
        - timestamp: whether the dataset version depends on the time of the download
    """

    name: Optional[str] = None
    """The name of the dataset"""

    url: Optional[str] = None
    """The URL of the dataset"""

    doi: Optional[str] = None
    """The DOI of this dataset"""

    def __init__(self, repository: Optional["Repository"]):
        super().__init__()
        self.repository = repository
        self.timestamp = False
        self.aliases = set()

        # Associated resources
        self.resources: Dict[str, "Download"] = {}
        self.ordered_resources = []

        # Hooks
        # pre-use: before returning the dataset object
        # pre-download: before downloading the dataset
        self.hooks = {"pre-use": [], "pre-download": []}

        self.url = None
        self.version = None

    @property
    def description(self):
        raise NotImplementedError(f"For class {self.__class__}")

    @property
    def configtype(self):
        raise NotImplementedError()

    @property
    def context(self):
        if self.repository is None:
            from datamaestro.context import Context  # noqa: F811

            return Context.instance()
        return self.repository.context

    def prepare(self, download=False) -> "Base":
        ds = self._prepare()
        ds.__datamaestro_dataset__ = self

        if download:
            ds.download()
        return ds

    def register_hook(self, hookname: str, hook: Callable):
        self.hooks[hookname].append(hook)

    @abstractmethod
    def _prepare(self) -> "Base": ...

    def format(self, encoder: str) -> str:
        s = self.prepare()
        if encoder == "normal":
            from .utils import JsonEncoder

            return JsonEncoder().encode(s)
        elif encoder == "xpm":
            from .utils import XPMEncoder

            return XPMEncoder().encode(s)
        else:
            raise Exception("Unhandled encoder: {encoder}")

    def setDataIDs(self, data: Config, id: str):
        """Set nested IDs automatically"""
        from datamaestro.data import Base

        if isinstance(data, Base):
            try:
                if data.id:
                    # There is already an ID, skip this
                    # and the descendants
                    return
            except KeyError:
                pass

            if self.repository is None:
                data.id = id
            else:
                data.id = f"{id}@{self.repository.name}"
        for key, value in data.__xpm__.values.items():
            if isinstance(value, Config):
                self.setDataIDs(value, f"{id}.{key}")

    def download(self, force=False):
        """Download all the necessary resources.

        Uses DAG-based topological ordering and the two-path system:
        1. Acquire exclusive lock (.state.lock)
        2. Resource writes to transient_path (under .downloads/)
        3. Framework moves transient_path → path (main folder)
        4. State marked COMPLETE
        5. Transient dependencies cleaned up eagerly
        6. .downloads/ directory removed after all resources complete
        7. Release lock
        """
        import fcntl

        from datamaestro.download import ResourceState

        self.prepare()
        logging.info(
            "Materializing %d resources",
            len(self.ordered_resources),
        )

        self.datapath.mkdir(parents=True, exist_ok=True)
        lock_path = self.datapath / ".state.lock"
        lock_file = lock_path.open("w")
        try:
            fcntl.flock(lock_file, fcntl.LOCK_EX)
            success = self._download_locked(force, ResourceState)
        finally:
            fcntl.flock(lock_file, fcntl.LOCK_UN)
            lock_file.close()

        return success

    def _download_locked(self, force, ResourceState):
        """Inner download logic, called while holding .state.lock."""
        success = True

        for resource in self.ordered_resources:
            # Step 1: Check state
            current_state = resource.state

            if current_state == ResourceState.COMPLETE and not force:
                # Verify files are actually present on disk
                if resource.has_files() and not resource.path.exists():
                    logging.warning(
                        "Resource %s marked COMPLETE but files "
                        "missing at %s — re-downloading",
                        resource.name,
                        resource.path,
                    )
                    resource.state = ResourceState.NONE
                    current_state = ResourceState.NONE
                else:
                    continue

            # Adopt pre-existing files (old downloads without state file)
            if (
                current_state == ResourceState.NONE
                and not force
                and resource.has_files()
                and resource.path.exists()
            ):
                logging.info(
                    "Resource %s already exists at %s — marking COMPLETE",
                    resource.name,
                    resource.path,
                )
                resource.state = ResourceState.COMPLETE
                continue

            if current_state == ResourceState.PARTIAL:
                if not resource.can_recover:
                    _delete_path(resource.transient_path)
                    resource.state = ResourceState.NONE

            # Verify all dependencies are COMPLETE
            for dep in resource.dependencies:
                if dep.state != ResourceState.COMPLETE:
                    logging.error(
                        "Dependency %s of %s is not COMPLETE",
                        dep.name,
                        resource.name,
                    )
                    return False

            # Step 2-4: Download with framework-managed state
            try:
                resource.download(force=force)

                # Move transient -> final, mark COMPLETE
                if resource.has_files():
                    _move_path(resource.transient_path, resource.path)
                resource.state = ResourceState.COMPLETE

            except Exception:
                logging.error("Could not download resource %s", resource)
                traceback.print_exc()

                # Handle PARTIAL state
                if resource.has_files() and resource.transient_path.exists():
                    if resource.can_recover:
                        resource.state = ResourceState.PARTIAL
                    else:
                        _delete_path(resource.transient_path)
                        resource.state = ResourceState.NONE
                success = False
                break

            # Step 5: Eager transient cleanup
            for dep in resource.dependencies:
                if dep.transient and all(
                    d.state == ResourceState.COMPLETE for d in dep.dependents
                ):
                    dep.cleanup()

        # Step 6: Remove .downloads/ directory after success
        if success:
            downloads_dir = self.datapath / ".downloads"
            if downloads_dir.is_dir():
                shutil.rmtree(downloads_dir)

        return success

    @staticmethod
    def find(name: str, context: Optional["Context"] = None) -> "DataDefinition":
        """Find a dataset given its name"""
        from datamaestro.context import Context  # noqa: F811

        context = Context.instance() if context is None else context

        logging.debug("Searching dataset %s", name)
        for repository in Context.instance().repositories():
            logging.debug("Searching dataset %s in %s", name, repository)
            dataset = repository.search(name)
            if dataset is not None:
                return dataset
        raise Exception("Could not find the dataset %s" % (name))


class FutureAttr:
    """Allows to access a dataset sub-property"""

    def __init__(self, dataset, keys):
        self.dataset = dataset
        self.keys = keys

    def __repr__(self):
        return "[%s].%s" % (self.dataset.id, ".".join(self.keys))

    def __call__(self):
        """Returns the value"""
        value = self.dataset.prepare()
        for key in self.keys:
            value = getattr(value, key)
        return value

    def __getattr__(self, key):
        return FutureAttr(self.dataset, self.keys + [key])

    def download(self, force=False):
        self.dataset.download(force)


class DatasetWrapper(AbstractDataset):
    """Wraps an annotated method into a dataset

    This is the standard way to define a dataset in datamaestro through
    annotations (otherwise, derive from `AbstractDataset`).
    """

    def __init__(self, annotation: "dataset", t: type):
        self.config = None
        self.repository: Optional[Repository] = None
        self.t = t
        self.base = annotation.base
        assert self.base is not None, f"Could not set the Config type for {t}"

        repository, components = DataDefinition.repository_relpath(t)
        super().__init__(repository)

        self.module_name = None
        if repository is None:
            # Try to find the module name
            self.module_name, _ = t.__module__.split(".", 1)

        # Set some variables
        self.url = annotation.url
        self.doi = annotation.doi
        self.as_prepare = annotation.as_prepare

        # Builds the ID:
        # Removes module_name.config prefix
        if (
            (annotation.id is None)
            or (annotation.id == "")
            or ("." not in annotation.id)
            or (annotation.id[0] == ".")
        ):
            # Computes an ID
            assert (
                # id is empty string = use the module id
                components[0] == "config"
            ), (
                "A @dataset without `id` should be in the "
                f".config module (not {t.__module__})"
            )

            if annotation.id is None:
                # There is nothing, use the full path
                path = ".".join(components[1:])
            else:
                # Replace
                path = ".".join(components[1:-1])
                if annotation.id != "":
                    path = f"{path}.{annotation.id}"

            self.id = path
        else:
            # Use the provided ID
            self.id = annotation.id

        self.aliases.add(self.id)

        # Get the documentation
        self._name = None
        self._description = None

    @property
    def name(self):
        self._process_doc()
        return self._name

    @property
    def description(self):
        self._process_doc()
        return self._description

    def _process_doc(self):
        if self._description is None:
            if self.t.__doc__:
                lines = self.t.__doc__.split("\n")
                self._name = lines[0]
                if len(lines) > 1:
                    assert lines[1].strip() == "", "Second line should be blank"
                if len(lines) > 2:
                    # Remove the common indent
                    lines = [line.rstrip() for line in lines[2:]]
                    minindent = max(
                        next(idx for idx, chr in enumerate(s) if not chr.isspace())
                        for s in lines
                        if len(s) > 0
                    )
                    self._description = "\n".join(
                        s[minindent:] if len(s) > 0 else "" for s in lines
                    )
            else:
                self._name = ""
                self._description = ""

    @property
    def configtype(self):
        return self.base

    def __getattr__(self, key):
        """Returns a pointer to a potential attribute"""
        return FutureAttr(self, [key])

    def download(self, force=False):
        if self.base is self.t:
            self._prepare()
        return super().download(force=force)

    def _prepare(self) -> "Base":
        if self.config is not None:
            return self.config

        # Direct creation of the dataset
        if self.base is self.t:
            self.config = self.base.__create_dataset__(self)

        else:
            # Construct the object
            if self.as_prepare:
                result = self.t(self, None)
            else:
                resources = {
                    key: value.prepare() for key, value in self.resources.items()
                }
                result = self.t(**resources)

            if result is None:
                raise RuntimeError(f"{self.base} did not return any resource")

            # Download resources
            logging.debug(
                "Building with data type %s and dataset %s", self.base, self.t
            )
            for hook in self.hooks["pre-use"]:
                hook(self)

            if result is None:
                name = self.t.__name__
                filename = inspect.getfile(self.t)
                raise Exception(
                    f"The dataset method {name} defined in "
                    f"{filename} returned a null object"
                )

            if isinstance(result, dict):
                self.config = self.base.C(**result)
            elif isinstance(result, self.base):
                self.config = result
            else:
                name = self.t.__name__
                filename = inspect.getfile(self.t)
                raise RuntimeError(
                    f"The dataset method {name} defined in "
                    f"{filename} returned an object of type {type(dict)}"
                )

        # Setup ourself
        self.config.__datamaestro_dataset__ = self

        # Set the ids
        self.setDataIDs(self.config, self.id)

        return self.config

    __call__ = _prepare

    @property
    def _path(self) -> Path:
        """Returns a unique relative path for this dataset"""
        path = Path(*self.id.split("."))
        if self.version:
            path = path.with_suffix(".v%s" % self.version)
        return path

    @property
    def datapath(self):
        """Returns the destination path for downloads"""
        if self.repository is not None:
            return self.repository.datapath / self._path

        # No repository, use __custom__/[MODULE NAME]
        path = self.context.datapath / "__custom__" / self.module_name / self._path

        return path

    def has_files(self) -> bool:
        """Returns whether this dataset has files or only includes references."""
        for resource in self.resources.values():
            if resource.has_files():
                return True
        return False

    def hasfiles(self) -> bool:
        """Deprecated: use has_files() instead."""
        import warnings

        warnings.warn(
            "hasfiles() is deprecated, use has_files()",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.has_files()


# --- Annotations

T = TypeVar("T")


class DataAnnotation:
    def __call__(self, object: T) -> T:
        if isinstance(object, AbstractDataset):
            self.annotate(object)
        else:
            if "__datamaestro__" in object.__dict__:
                self.annotate(object.__datamaestro__)
            else:
                # With configuration objects, add a __datamaestro__ member to the class
                assert issubclass(object, Config), (
                    f"{object} cannot be annotated (only dataset or data definitions)"
                )
                if "__datamaestro__" not in object.__dict__:
                    object.__datamaestro__ = AbstractData()
                self.annotate(object.__datamaestro__)

        return object

    def annotate(self, data: AbstractData):
        raise NotImplementedError("Method annotate for class %s" % self.__class__)


class DatasetAnnotation:
    """Base class for all annotations"""

    def __call__(self, dataset: AbstractDataset):
        if isinstance(dataset, AbstractDataset):
            self.annotate(dataset)
        elif issubclass(dataset, Dataset):
            self.annotate(dataset.__datamaestro__)
        else:
            raise RuntimeError(
                f"Only datasets can be annotated with {self}, "
                f"but {dataset} is not a dataset"
            )

        return dataset

    def annotate(self, dataset: AbstractDataset):
        raise NotImplementedError("Method annotate for class %s" % self.__class__)


def hook(name: str):
    """Annotate a method of a DatasetAnnotation class to be a hook"""

    class HookAnnotation(DatasetAnnotation):
        def __init__(self, callable: Callable, args, kwargs):
            self.callable = callable
            self.args = args
            self.kwargs = kwargs

        def annotate(self, dataset):
            dataset.register_hook(name, self._hook)

        def _hook(self, definition: DataDefinition):
            self.callable(definition, *self.args, **self.kwargs)

    def annotate(callable: Callable):
        def collect(*args, **kwargs):
            return HookAnnotation(callable, args, kwargs)

        return collect

    return annotate


def DataTagging(f):
    class Annotation(DataAnnotation):
        """Define tags in a data definition"""

        def __init__(self, *tags):
            self.tags = tags

        def annotate(self, metadata):
            f(metadata).update(self.tags)

    return Annotation


datatags = DataTagging(lambda d: d.tags)
datatasks = DataTagging(lambda d: d.tasks)


class metadata:
    def __init__(
        self,
        tags: Union[str, List[str]] = None,
        tasks: Union[str, List[str]] = None,
    ):
        pass

    def __call__(self, object: type):
        # FIXME: todo
        return object


class dataset:
    """Dataset decorator

    Meta-datasets are not associated with any base type.

    :param base: The base type (or None if inferred from type annotation).
    :param timestamp: If the dataset evolves, specify its timestamp.
    :param id: Gives the full ID of the dataset if it contains a '.',
        the last component if not containing a '.', or the last components
        if starting with '.'
    :param url: The URL associated with the dataset.
    :param size: The size of the dataset (should be a parsable format).
    :param doi: The DOI of the corresponding paper.
    :param as_prepare: Resources are setup within the method itself
    """

    def __init__(
        self,
        base=None,
        *,
        timestamp: str | None = None,
        id: None | str = None,
        url: None | str = None,
        size: None | int | str = None,
        doi: None | str = None,
        as_prepare: bool = False,
    ):
        if hasattr(base, "__datamaestro__") and isinstance(
            base.__datamaestro__, metadataset
        ):
            self.base = base.__datamaestro__.base
        else:
            self.base = base

        self.id = id
        self.url = url
        self.meta = False
        self.timestamp = timestamp
        self.size = size
        self.doi = doi
        self.as_prepare = as_prepare

    def __call__(self, t):
        from datamaestro.data import Base

        try:
            if self.base is None:
                if inspect.isclass(t) and issubclass(t, Base):
                    self.base = t
                else:
                    try:
                        # Get type from return annotation
                        return_type = t.__annotations__["return"]
                        if isinstance(return_type, _GenericAlias):
                            return_type = return_type.__origin__
                        self.base = return_type
                    except KeyError:
                        logging.warning("No return annotation in %s", t)
                        raise
            object.__getattribute__(t, "__datamaestro__")
            raise AssertionError("@data should only be called once")
        except AttributeError:
            pass
        dw = DatasetWrapper(self, t)
        t.__dataset__ = dw

        # For class-based datasets, scan for Resource class attributes
        if inspect.isclass(t) and issubclass(t, Base):
            _bind_class_resources(t, dw)
            return t
        return dw


class metadataset(AbstractDataset):
    """Annotation for object/functions which are abstract dataset definitions

    i.e. shared by more than one real dataset. This is useful to share tags,
    urls, etc.
    """

    def __init__(self, base):
        super().__init__(None)
        self.base = base

    def __call__(self, t):
        try:
            object.__getattribute__(t, "__datamaestro__")
            raise AssertionError("@data should only be called once")
        except AttributeError:
            pass
        t.__datamaestro__ = self
        return t

    _prepare = None



"""Single file download resources.

Provides FileResource subclasses for downloading individual files
from URLs, with optional transforms and integrity checking.
"""

from __future__ import annotations

import io
import gzip
import logging
import os
import os.path as op
import shutil
import tarfile
from pathlib import Path

import urllib3

from datamaestro.download import FileResource
from datamaestro.stream import Transform
from datamaestro.utils import copyfileobjs

logger = logging.getLogger(__name__)


def open_ext(*args, **kwargs):
    """Opens a file according to its extension."""
    name = args[0]
    if name.endswith(".gz"):
        return gzip.open(*args, *kwargs)
    return io.open(*args, **kwargs)


class FileDownloader(FileResource):
    """Downloads a single file from a URL.

    Supports optional transforms (e.g., gzip decompression)
    and integrity checking.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = FileDownloader.apply(
                "data.csv", "http://example.com/data.csv.gz"
            )

    Usage as decorator (deprecated)::

        @filedownloader("data.csv", "http://example.com/data.csv.gz")
        @dataset(Base)
        def my_dataset(data): ...
    """

    def __init__(
        self,
        filename: str,
        url: str,
        size: int | None = None,
        transforms: Transform | None = None,
        checker=None,
        *,
        varname: str | None = None,
        transient: bool = False,
    ):
        """
        Args:
            filename: The filename within the data folder; the variable
                name corresponds to the filename without the extension.
            url: The URL to download.
            size: Expected size in bytes (or None).
            transforms: Transform the file before storing it.
                Auto-detected from URL path if None.
            checker: File integrity checker.
            varname: Explicit resource name.
            transient: If True, data can be deleted after dependents
                complete.
        """
        super().__init__(filename, varname=varname, transient=transient)
        self.url = url
        self.checker = checker
        self.size = size

        p = urllib3.util.parse_url(self.url)
        path = Path(Path(p.path).name)
        self.transforms = transforms if transforms else Transform.createFromPath(path)

    def _download(self, destination: Path) -> None:
        logger.info("Downloading %s into %s", self.url, destination)

        # Creates directory if needed
        dir = op.dirname(destination)
        os.makedirs(dir, exist_ok=True)

        # Download (cache)
        with self.context.downloadURL(self.url, size=self.size) as file:
            # Transform if need be
            if self.transforms:
                logger.info("Transforming file")
                with (
                    self.transforms(file.path.open("rb")) as stream,
                    destination.open("wb") as out,
                ):
                    if self.checker:
                        copyfileobjs(stream, [out, self.checker])
                        self.checker.close()
                    else:
                        shutil.copyfileobj(stream, out)
            else:
                logger.info("Keeping original downloaded file %s", file.path)
                if self.checker:
                    self.checker.check(file.path)
                (shutil.copy if file.keep else shutil.move)(file.path, destination)

        logger.info("Created file %s", destination)


# Factory alias for backward compat and convenient usage
filedownloader = FileDownloader.apply


class ConcatDownloader(FileResource):
    """Concatenate all files from an archive into a single file.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = ConcatDownloader.apply(
                "data.txt", "http://example.com/data.tar.gz"
            )
    """

    def __init__(
        self,
        filename: str,
        url: str,
        transforms=None,
        *,
        varname: str | None = None,
        transient: bool = False,
    ):
        """
        Args:
            filename: The filename within the data folder; the variable
                name corresponds to the filename without the extension.
            url: The URL to download.
            transforms: Transform the file before storing it.
            varname: Explicit resource name.
            transient: If True, data can be deleted after dependents
                complete.
        """
        super().__init__(filename, varname=varname, transient=transient)
        self.url = url
        self.transforms = transforms

    def _download(self, destination: Path) -> None:
        with (
            self.context.downloadURL(self.url) as dl,
            tarfile.open(dl.path) as archive,
        ):
            destination.parent.mkdir(parents=True, exist_ok=True)

            with open(destination, "wb") as out:
                for tarinfo in archive:
                    if tarinfo.isreg():
                        transforms = self.transforms or Transform.createFromPath(
                            Path(tarinfo.name)
                        )
                        logger.debug("Processing file %s", tarinfo.name)
                        with transforms(archive.fileobject(archive, tarinfo)) as fp:
                            shutil.copyfileobj(fp, out)


# Factory alias for backward compat
concatdownload = ConcatDownloader.apply


# --- Backward compat aliases ---
# Keep old class names importable but they now point to new classes

SingleDownload = FileDownloader



from datamaestro.download import Resource


class Todo(Resource):
    """Placeholder resource indicating download is not yet implemented."""

    def download(self, force=False):
        raise NotImplementedError(
            "Download method not defined - please edit the definition file"
        )

    def prepare(self):
        raise NotImplementedError(
            "Prepare method not defined - please edit the definition file"
        )



"""Resource system for dataset download and processing pipelines.

This module defines the Resource interface and its concrete subclasses
(FileResource, FolderResource, ValueResource) for managing dataset
download and preprocessing steps as a directed acyclic graph (DAG).
"""

from __future__ import annotations

import json
import logging
import shutil
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import IO, Union

from attrs import define

from datamaestro.definitions import AbstractDataset, DatasetAnnotation
from datamaestro.utils import deprecated

logger = logging.getLogger(__name__)

# Module-level deprecation tracking (emit each category only once)
_deprecation_warned: set[str] = set()


def _warn_once(category: str, message: str):
    """Emit a deprecation warning only once per category."""
    if category not in _deprecation_warned:
        _deprecation_warned.add(category)
        warnings.warn(message, DeprecationWarning, stacklevel=3)


# --- State metadata file helpers ---


class ResourceStateFile:
    """Manages the .state.json metadata file for resource states.

    Location: <dataset.datapath>/.state.json

    Format:
        {
          "version": 1,
          "resources": {
            "RESOURCE_NAME": {"state": "none"|"partial"|"complete"},
            ...
          }
        }
    """

    VERSION = 1

    def __init__(self, datapath: Path):
        self._path = datapath / ".state.json"

    def read(self, resource_name: str) -> "ResourceState":
        """Read the state for a resource. Returns NONE if not found."""
        data = self._load()
        entry = data.get("resources", {}).get(resource_name)
        if entry is None:
            return ResourceState.NONE
        return ResourceState(entry["state"])

    def write(self, resource_name: str, state: "ResourceState"):
        """Write the state for a resource (atomic write)."""
        data = self._load()
        if "resources" not in data:
            data["resources"] = {}
        data["resources"][resource_name] = {"state": state.value}
        self._save(data)

    def _load(self) -> dict:
        if self._path.is_file():
            with self._path.open("r") as f:
                return json.load(f)
        return {"version": self.VERSION, "resources": {}}

    def _save(self, data: dict):
        self._path.parent.mkdir(parents=True, exist_ok=True)
        tmp = self._path.with_suffix(".tmp")
        with tmp.open("w") as f:
            json.dump(data, f, indent=2)
        tmp.rename(self._path)


# --- ResourceState enum ---


class ResourceState(str, Enum):
    """State of a resource in the preparation pipeline."""

    NONE = "none"
    """Not started / no data on disk."""

    PARTIAL = "partial"
    """Started but incomplete (error during download)."""

    COMPLETE = "complete"
    """Fully available."""


# --- Lazy initialization decorator (backward compat) ---


def initialized(method):
    """Ensure the object is initialized (calls postinit on first use).

    Deprecated: new Resource subclasses should not rely on this pattern.
    """

    def wrapper(self, *args, **kwargs):
        if not self._post:
            self._post = True
            self.postinit()
        return method(self, *args, **kwargs)

    return wrapper


# --- SetupOptions (backward compat) ---


@define(kw_only=True)
class SetupOptions:
    pass


# --- Resource base class ---


class Resource(DatasetAnnotation, ABC):
    """Base class for all dataset resources.

    A resource represents a single step in a dataset preparation pipeline.
    Resources form a DAG: each resource declares its dependencies, and
    the orchestrator ensures they are processed in topological order.

    Usage modes:

    1. Class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = filedownloader("data.csv", "http://...", transient=True)
            PROCESSED = SomeProcessor.from_file(DATA)

    2. Decorator on function (deprecated, backward compat)::

        @filedownloader("data.csv", "http://...")
        @dataset(Base)
        def my_dataset(data): ...

    Two-path system:

    - ``transient_path``: where download/processing writes data
    - ``path``: final location after successful completion

    The framework moves data from ``transient_path`` → ``path`` and then
    marks the resource as COMPLETE. Subclass ``download()`` implementations
    should always write to ``transient_path``.

    State is persisted in a metadata file at::

        <dataset.datapath>/.downloads/.state.json
    """

    def __init__(
        self,
        varname: str | None = None,
        *,
        transient: bool = False,
    ):
        """
        Args:
            varname: Explicit resource name. If None, auto-set from
                class attribute name during binding. Required when
                used as a decorator (backward compat mode).
            transient: If True, this resource's data can be deleted
                after all its dependents reach COMPLETE.
        """
        self.name: str | None = varname
        self._name_explicit: bool = varname is not None
        self.dataset: AbstractDataset | None = None
        self.transient: bool = transient
        self._dependencies: list[Resource] = []
        self._dependents: list[Resource] = []

        # Backward compat: lazy initialization support
        self._post = False

    # ---- Properties ----

    @property
    def can_recover(self) -> bool:
        """Whether partial downloads can be resumed.

        When True and state is PARTIAL, existing data at transient_path
        is preserved on error, allowing the next download() call to
        resume from where it left off.

        When False and state is PARTIAL, data at transient_path is
        deleted and state is reset to NONE.

        Default: False. Subclasses override to enable recovery.
        """
        return False

    @property
    def dependencies(self) -> list[Resource]:
        """Resources that must be COMPLETE before this one can process.

        Populated from constructor arguments. Subclasses with factory
        methods should pass dependency resources to ``__init__`` and
        store them in ``_dependencies``.
        """
        return self._dependencies

    @property
    def dependents(self) -> list[Resource]:
        """Resources that depend on this one (inverse of dependencies).

        Computed by the dataset after all resources are bound.
        Used for eager transient cleanup decisions.
        """
        return self._dependents

    @property
    def path(self) -> Path:
        """Final storage path for this resource's data.

        This is where data lives after successful completion.
        Default: ``dataset.datapath / self.name``

        Subclasses may override to customize (e.g., add file extension).
        """
        return self.dataset.datapath / self.name

    @property
    def transient_path(self) -> Path:
        """Temporary path where download/processing writes data.

        During download(), subclasses write to this path.
        After successful download, the framework moves the data from
        transient_path to path, then marks state as COMPLETE.

        Default: ``dataset.datapath / ".downloads" / self.name``
        """
        return self.dataset.datapath / ".downloads" / self.name

    @property
    def state(self) -> ResourceState:
        """Current state, read from the metadata file.

        If no metadata entry exists, returns NONE.
        """
        if self.dataset is None:
            return ResourceState.NONE
        state_file = ResourceStateFile(self.dataset.datapath)
        return state_file.read(self.name)

    @state.setter
    def state(self, value: ResourceState) -> None:
        """Update state in the metadata file (atomic write)."""
        state_file = ResourceStateFile(self.dataset.datapath)
        state_file.write(self.name, value)

    @property
    def context(self):
        """Application context (from dataset)."""
        return self.dataset.context

    # ---- Abstract methods ----

    @abstractmethod
    def download(self, force: bool = False) -> None:
        """Execute this resource's download/processing step.

        Contract:

        - Called only when all dependencies are COMPLETE.
        - Must write output to ``self.transient_path``.
        - The framework handles moving transient_path → path
          and setting state to COMPLETE after this returns.
        - If force=True, re-execute even if already COMPLETE.

        Note: State management (COMPLETE/PARTIAL/NONE transitions,
        moving transient_path → path) is handled by the framework,
        NOT by the download() implementation.

        Raises:
            Exception: On download/processing failure. The framework
                will handle PARTIAL state based on can_recover.
        """
        ...

    @abstractmethod
    def prepare(self):
        """Return the value for dataset construction.

        Called after download() has completed (state is COMPLETE).
        Return type depends on the resource subclass:

        - FileResource → Path
        - FolderResource → Path
        - ValueResource → resource-specific

        For backward compat with function-based datasets, this value
        is passed as a keyword argument to the dataset function.
        """
        ...

    # ---- Concrete methods ----

    def cleanup(self) -> None:
        """Remove this resource's data from disk.

        Called automatically for transient resources after all
        dependents reach COMPLETE (eager cleanup).

        Default implementation:

        - Deletes self.path (file or directory)
        - Deletes self.transient_path if it exists
        - Sets self.state = NONE

        Subclasses may override for custom cleanup.
        """
        for p in (self.path, self.transient_path):
            if p.exists():
                if p.is_dir():
                    shutil.rmtree(p)
                else:
                    p.unlink()
        self.state = ResourceState.NONE

    def has_files(self) -> bool:
        """Whether this resource produces files on disk.

        Returns False for reference-only resources (e.g., links
        to other datasets, in-memory values).
        Default: True.
        """
        return True

    # Backward compat alias
    def hasfiles(self) -> bool:
        """Deprecated: use has_files() instead."""
        _warn_once("hasfiles", "hasfiles() is deprecated, use has_files()")
        return self.has_files()

    def postinit(self):
        """Legacy lazy initialization hook.

        Deprecated: new Resource subclasses should perform
        initialization in ``__init__`` or ``bind()``.
        """
        pass

    # ---- Binding ----

    def bind(self, name: str, dataset: AbstractDataset) -> None:
        """Bind this resource to a dataset.

        Called by the dataset class machinery during initialization.
        Sets self.name (if not explicitly set via varname) and
        self.dataset. Registers the resource in dataset.resources
        and dataset.ordered_resources.

        For class-based datasets: called by ``@dataset`` when it
        processes class attributes.
        For decorator-based: called by ``annotate()`` (existing protocol).
        """
        if not self._name_explicit:
            self.name = name

        assert self.dataset is None, (
            f"Resource {self.name} is already bound to a dataset"
        )

        if self.name in dataset.resources:
            raise AssertionError(f"Name {self.name} already declared as a resource")

        dataset.resources[self.name] = self
        dataset.ordered_resources.append(self)
        self.dataset = dataset

    def annotate(self, dataset: AbstractDataset) -> None:
        """Register with a dataset (DatasetAnnotation protocol).

        Deprecated for new code. Calls bind() internally.
        """
        _warn_once(
            "annotate",
            "Using resources as decorators is deprecated. "
            "Define them as class attributes instead.",
        )
        self.bind(self.name, dataset)

    def contextualize(self):
        """When using an annotation inline, uses the current
        dataset wrapper object.

        Deprecated: use class-attribute resource definitions instead.
        """
        wrapper = AbstractDataset.processing()
        self.annotate(wrapper)

    def setup(
        self,
        dataset: Union[AbstractDataset],
        options: SetupOptions = None,
    ):
        """Direct way to setup the resource (no annotation).

        Deprecated: use class-attribute resource definitions instead.
        """
        self(dataset)
        return self.prepare()

    # ---- Factory pattern ----

    @classmethod
    def apply(cls, *args, **kwargs) -> "Resource":
        """Factory classmethod for creating resource instances.

        Allows defining shorthand factory functions::

            filedownloader = FileDownloader.apply

        Default implementation: ``return cls(*args, **kwargs)``
        Subclasses may override for custom argument handling.
        """
        return cls(*args, **kwargs)

    # ---- Backward compat: definition property ----

    @property
    def definition(self) -> AbstractDataset | None:
        """Deprecated: use ``dataset`` attribute instead."""
        _warn_once(
            "definition",
            "Resource.definition is deprecated, use Resource.dataset",
        )
        return self.dataset

    # Backward compat: varname property
    @property
    def varname(self) -> str | None:
        """Deprecated: use ``name`` attribute instead."""
        _warn_once(
            "varname",
            "Resource.varname is deprecated, use Resource.name",
        )
        return self.name

    @varname.setter
    def varname(self, value: str | None):
        self.name = value


# --- FileResource ---


class FileResource(Resource):
    """A resource that produces a single file on disk.

    Subclasses implement ``_download()`` to produce the file at the
    given destination (which is ``self.transient_path``).
    """

    def __init__(
        self,
        filename: str,
        *,
        varname: str | None = None,
        transient: bool = False,
    ):
        """
        Args:
            filename: The filename (with extension) for the produced file.
                Used to construct the storage path.
            varname: Explicit resource name. If None, derived from
                filename (extension stripped) or class attribute name.
            transient: See Resource.
        """
        import re

        effective_varname = varname or re.sub(r"\..*$", "", filename)
        super().__init__(varname=effective_varname, transient=transient)
        # Only mark name as explicit if user actually passed varname
        self._name_explicit = varname is not None
        self.filename = filename

    @property
    def path(self) -> Path:
        """Final path to the produced file.

        ``dataset.datapath / self.filename``
        """
        return self.dataset.datapath / self.filename

    @property
    def transient_path(self) -> Path:
        """Temporary path for writing during download.

        ``dataset.datapath / ".downloads" / self.filename``
        """
        return self.dataset.datapath / ".downloads" / self.filename

    def prepare(self) -> Path:
        """Returns self.path."""
        return self.path

    def stream(self) -> IO[bytes] | None:
        """Return a readable byte stream of the file content.

        Returns None if streaming is not supported for this resource.
        Default: returns None. Subclasses may override.

        This allows downstream resources to consume data without
        needing the file to be fully materialized on disk first.
        """
        return None

    def download(self, force: bool = False) -> None:
        """Downloads the file.

        Delegates to ``_download(self.transient_path)``.
        """
        self._download(self.transient_path)

    @abstractmethod
    def _download(self, destination: Path) -> None:
        """Subclass hook: download/produce the file at destination.

        Args:
            destination: The path to write the file to
                (``self.transient_path``).
        """
        ...


# --- FolderResource ---


class FolderResource(Resource):
    """A resource that produces a directory on disk.

    Subclasses implement ``_download()`` to populate the directory at
    the given destination (which is ``self.transient_path``).
    """

    @property
    def path(self) -> Path:
        """Final path to the produced directory.

        ``dataset.datapath / self.name``
        """
        return self.dataset.datapath / self.name

    @property
    def transient_path(self) -> Path:
        """Temporary path for writing during download.

        ``dataset.datapath / ".downloads" / self.name``
        """
        return self.dataset.datapath / ".downloads" / self.name

    def prepare(self) -> Path:
        """Returns self.path."""
        return self.path

    def download(self, force: bool = False) -> None:
        """Downloads/extracts the directory content to transient_path."""
        self._download(self.transient_path)

    @abstractmethod
    def _download(self, destination: Path) -> None:
        """Subclass hook: populate the directory at destination.

        Args:
            destination: The path to write to (``self.transient_path``).
        """
        ...


# --- ValueResource ---


class ValueResource(Resource):
    """A resource that produces an in-memory value (no files on disk).

    Used for resources like HuggingFace dataset handles that don't
    produce local files. The transient_path/path two-path system
    is not used; state tracking is still via metadata file.
    """

    def has_files(self) -> bool:
        return False

    @abstractmethod
    def prepare(self):
        """Return the in-memory value."""
        ...


# --- Deprecated compatibility classes ---


class Download(Resource):
    """Deprecated: use Resource instead."""

    def __init_subclass__(cls):
        _warn_once(
            f"Download-{cls.__name__}",
            f"Download is deprecated ({cls}): use `Resource`",
        )
        return super().__init_subclass__()


# --- reference resource ---


class reference(Resource):
    """References another dataset instead of downloading."""

    def __init__(self, varname=None, reference=None):
        """
        Args:
            varname: The name of the variable.
            reference: Another dataset to reference.
        """
        super().__init__(varname=varname)
        assert reference is not None, "Reference cannot be null"
        self.reference = reference

    def prepare(self):
        v = self.reference.prepare()
        if isinstance(v, AbstractDataset):
            return v().prepare()
        return v

    def download(self, force=False):
        self.reference.__datamaestro__.download(force)

    def has_files(self):
        # We don't really have files
        return False


Reference = deprecated("Use @reference instead of @Reference", reference)



#!/usr/bin/env python3
# ruff: noqa: T201

from importlib.metadata import entry_points
import sys
import logging
from functools import update_wrapper
import traceback as tb
import re
from pathlib import Path
import shutil
from typing import Set
from urllib.parse import urlparse

import click

import datamaestro
from .context import Context

logging.basicConfig(level=logging.INFO)


class Config:
    def __init__(self, context: Context):
        self.context = context
        self.traceback = False
        self.host = None


def pass_cfg(f):
    """Pass configuration information"""

    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(f, ctx.obj, *args, **kwargs)

    return update_wrapper(new_func, f)


# Get all the available repositories

REPOSITORIES = {}
for entry_point in entry_points(group="datamaestro.repositories"):
    REPOSITORIES[entry_point.name] = entry_point


# --- Create the argument parser


@click.group()
@click.option("--quiet", is_flag=True, help="Be quiet")
@click.option("--keep-downloads", is_flag=True, help="Keep downloads")
@click.option("--debug", is_flag=True, help="Be even more verbose (implies traceback)")
@click.option("--host", type=str, help="Remote hostname", default=None)
@click.option(
    "--pythonpath",
    type=str,
    help="Remote python path (default python)",
    default="python",
)
@click.option(
    "--traceback", is_flag=True, help="Display traceback if an exception occurs"
)
@click.option(
    "--data",
    type=Path,
    help="Directory containing datasets",
    default=Context.MAINDIR,
)
@click.pass_context
def cli(ctx, quiet, debug, traceback, data, keep_downloads, host, pythonpath):
    if quiet:
        logging.getLogger().setLevel(logging.WARN)
    elif debug:
        logging.getLogger().setLevel(logging.DEBUG)

    if host:
        context = Context.remote(host, pythonpath, data)
    else:
        context = Context(data)

    context.keep_downloads = keep_downloads
    context.traceback = traceback

    ctx.obj = Config(context)
    ctx.obj.traceback = traceback
    ctx.obj.debug = debug
    ctx.obj.host = host


def main():
    cli(obj=None)


@cli.command(help="Prints the full information about a dataset")
@click.argument("dataset", type=str)
@pass_cfg
def info(config: Config, dataset):
    from datamaestro.definitions import AbstractDataset

    dataset = AbstractDataset.find(dataset)
    print(dataset.name)
    if dataset.url:
        print(dataset.url)

    print(
        "Types (ids):",
        ", ".join(str(s.__getxpmtype__().identifier) for s in dataset.ancestors()),
    )
    print(
        "Types (class):",
        ", ".join(
            str(s.__module__ + "." + s.__qualname__) for s in dataset.ancestors()
        ),
    )
    if dataset.tags:
        print("Tags:", ", ".join(dataset.tags))
    if dataset.tasks:
        print("Tasks:", ", ".join(dataset.tasks))

    print()
    if dataset.description:
        print(dataset.description)
    else:
        print("(no description provided)")


# --- General information


@cli.command(help="List available repositories")
def repositories():
    for name, entry_point in REPOSITORIES.items():
        repo_class = entry_point.load()
        print("%s: %s" % (entry_point.name, repo_class.DESCRIPTION))


@cli.command(help="Get version")
def version():
    print(datamaestro.__version__)


# --- Cleanup


@click.option("--size", is_flag=True, help="Show size")
@cli.command(help="List (and remove) orphan directories")
@pass_cfg
def orphans(config: Config, size):
    import subprocess

    for repository in config.context.repositories():
        # For each repository
        paths = set()
        ancestors: Set[Path] = set()
        for dataset in repository:
            if dataset.hasfiles():
                paths.add(dataset.datapath)
                ancestor = dataset.datapath.parent
                while ancestor not in ancestors:
                    ancestors.add(ancestor)
                    ancestor = ancestor.parent

        def lookup(path, prefix=[]):
            if path in paths:
                return
            if path not in ancestors:
                if size:
                    print(
                        subprocess.check_output(["du", "-hs", path.absolute()])
                        .decode("utf-8")
                        .strip(),
                    )
                else:
                    print(path)
                return

            for child in path.iterdir():
                if child.is_dir():
                    lookup(child, prefix + [child.name])
                else:
                    return True

        lookup(repository.datapath)


# --- Manage external data folders


@cli.group(help="Manage external data folders")
def datafolders():
    pass


@datafolders.command("list", help="List of external data folders")
@pass_cfg
def datafolder_list(config: Config):
    for key, value in config.context.settings.datafolders.items():
        print("%s\t%s" % (key, value))


@click.argument("path", type=Path)
@click.argument("key", type=str)
@datafolders.command("set", help="List of external data folders")
@pass_cfg
def datafolder_set(config: Config, key: str, path: Path):
    settings = config.context.settings
    settings.datafolders[key] = path
    settings.save()


# --- Create a dataset

DATASET_REGEX = re.compile(r"^\w[\w\.-]+\w$")


def dataset_id_check(ctx, param, value):
    try:
        url = urlparse(value)
        if url.scheme:
            return ".".join(url.hostname.split(".")[::-1] + url.path[1:].split("/"))
    except:
        raise

    if not DATASET_REGEX.match(value):
        raise click.BadParameter(
            "Dataset ID needs to be an URL or in the format AAA.BBBB.CCC[.DDD]"
        )
    return value


@click.argument("dataset-id", callback=dataset_id_check)
@click.argument("repository-id", type=click.Choice(REPOSITORIES.keys()))
@cli.command()
@pass_cfg
def create_dataset(config: Config, repository_id: str, dataset_id: str):
    """Create a new dataset in the repository repository-id"""
    # Construct the path of the new dataset definition
    repo_class = REPOSITORIES[repository_id].load()
    path = repo_class(config).configdir  # type: Path
    names = dataset_id.split(".")
    for name in names:
        path = path / name
    path = path.with_suffix(".py")

    if path.is_file():
        print("File {} already exists - not overwritting".format(path))
        sys.exit(1)

    template_path = Path(__file__).parent / "templates" / "dataset.py"
    path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(template_path, path)
    print("Created file {}".format(path))


# --- prepare and download


@click.argument("dataset")
@cli.command()
@pass_cfg
def download(config: Config, dataset):
    """Download a dataset"""
    from datamaestro.definitions import AbstractDataset

    dataset = AbstractDataset.find(dataset)
    success = dataset.download()
    if not success:
        logging.error("One or more errors occured while downloading the dataset")
        sys.exit(1)


@click.argument("datasetid")
@click.option(
    "--encoder",
    help="Encoder used for output",
    default="normal",
    type=click.Choice(["normal", "xpm"]),
)
@click.option("--no-downloads", is_flag=True, help="Do not try to download datasets")
@cli.command(help="Downloads a dataset (if freely available)")
@pass_cfg
def prepare(config: Config, datasetid, encoder, no_downloads):
    """Download a dataset and returns information in json format"""
    dataset = config.context.dataset(datasetid)

    if not no_downloads:
        success = dataset.download()
        if not success:
            logging.error("One or more errors occured while downloading the dataset")
            sys.exit(1)

    try:
        print(dataset.format(encoder))
    except Exception as e:
        if config.traceback:
            tb.print_exc()
        logging.error("Error encoding to JSON: %s", e)
        sys.exit(1)


# --- Search


@click.argument("searchterms", nargs=-1)  # , description="Search terms (e.g. tag:XXX)")
@cli.command(help="Search for a dataset")
@pass_cfg
def search(config: Config, searchterms):
    """Search for a dataset

    Repositories can be searched with repo:REGEX expression
    Tags can be searched with tag:REGEX expression
    Tasks can be searched with task:REGEX expression

    When no search modifier `MODIFIER:` is given, matches the dataset

    """
    from .search import Condition, AndCondition

    condition = AndCondition()
    for searchterm in searchterms:
        condition.append(Condition.parse(searchterm))

    logging.debug("Search: %s", condition)
    for dataset in config.context.datasets():
        try:
            if condition.match(dataset):
                cfg = dataset.configtype
                print(
                    "[%s] %s (%s)"
                    % (
                        dataset.repository.id,
                        dataset.id,
                        cfg.__name__ if cfg is not None else "?",
                    )
                )
        except Exception:
            logging.error("Error while matching with dataset %s", dataset)
            raise



from . import Transform


class Gunzip(Transform):
    def __call__(self, fileobj):
        import gzip

        return gzip.GzipFile(fileobj=fileobj)

    def path(self, path):
        if path.suffix == ".gz":
            return path.stem
        return path



"""Archive download resources.

Provides FolderResource subclasses for downloading and extracting
ZIP and TAR archives.
"""

from __future__ import annotations

import logging
import re
import shutil
import tarfile
import zipfile
from pathlib import Path
from typing import Set

import urllib3

from datamaestro.download import FolderResource
from datamaestro.utils import CachedFile, FileChecker

logger = logging.getLogger(__name__)


class ArchiveDownloader(FolderResource):
    """Abstract base for all archive-related extractors.

    Usage as class attribute (preferred)::

        @dataset(url="...")
        class MyDataset(Base):
            DATA = ZipDownloader.apply(
                "archive", "http://example.com/data.zip"
            )

    Usage as decorator (deprecated)::

        @zipdownloader("archive", "http://example.com/data.zip")
        @dataset(Base)
        def my_dataset(archive): ...
    """

    def __init__(
        self,
        varname: str,
        url: str,
        subpath: str | None = None,
        checker: FileChecker | None = None,
        files: Set[str] | None = None,
        *,
        transient: bool = False,
    ):
        """Downloads and extract the content of the archive.

        Args:
            varname: The name of the variable when defining the dataset.
            url: The archive URL.
            subpath: A subpath in the archive; only files from this
                subpath will be extracted.
            checker: The hash check for the downloaded file.
            files: A set of files; if present, only extract those.
            transient: If True, data can be deleted after dependents
                complete.
        """
        super().__init__(varname=varname, transient=transient)
        self.url = url
        self.subpath = subpath
        self.checker = checker
        self._files = set(files) if files else None
        if self.subpath and not self.subpath.endswith("/"):
            self.subpath = self.subpath + "/"

    def postinit(self):
        # Define the path
        p = urllib3.util.parse_url(self.url)
        self._archive_name = self._name(Path(p.path).name)

    @property
    def path(self) -> Path:
        """Final path to the extracted directory."""
        if not self._post:
            self._post = True
            self.postinit()

        if len(self.dataset.resources) > 1:
            return self.dataset.datapath / self._archive_name
        return self.dataset.datapath

    @property
    def transient_path(self) -> Path:
        """Temporary path for extraction."""
        if not self._post:
            self._post = True
            self.postinit()

        if len(self.dataset.resources) > 1:
            return self.dataset.datapath / ".downloads" / self._archive_name
        return self.dataset.datapath / ".downloads" / self.name

    @property
    def extractall(self):
        """Returns whether everything can be extracted."""
        return self._files is None and self.subpath is None

    def filter(self, iterable, getname):
        L = len(self.subpath) if self.subpath else 0

        for info in iterable:
            name = getname(info)
            logger.debug("Looking at %s", name)
            if self._files and name not in self._files:
                continue

            if self.subpath and name.startswith(self.subpath):
                yield info, name[L:]

            if not self.subpath:
                yield info, name

    def _download(self, destination: Path) -> None:
        logger.info("Downloading %s into %s", self.url, destination)

        destination.parent.mkdir(parents=True, exist_ok=True)

        with self.context.downloadURL(self.url) as file:
            if self.checker:
                self.checker.check(file.path)
            self.unarchive(file, destination)

        # Look at the content - if single directory, unwrap
        children = list(destination.iterdir())
        if len(children) == 1 and children[0].is_dir():
            single_dir = children[0]
            logger.info(
                "Moving single directory %s into destination %s",
                single_dir,
                destination,
            )
            # Move contents up one level
            tmp = destination.with_suffix(".unwrap")
            shutil.move(str(single_dir), str(tmp))
            shutil.rmtree(destination)
            shutil.move(str(tmp), str(destination))

    def unarchive(self, file, destination: Path):
        raise NotImplementedError()

    def _name(self, name: str) -> str:
        raise NotImplementedError()


class ZipDownloader(ArchiveDownloader):
    """ZIP Archive handler."""

    def _name(self, name):
        return re.sub(r"\.zip$", "", name)

    def unarchive(self, file, destination: Path):
        logger.info("Unzipping file")
        with zipfile.ZipFile(file.path) as zip:
            if self.extractall:
                zip.extractall(destination)
            else:
                for zip_info, name in self.filter(
                    zip.infolist(),
                    lambda zip_info: zip_info.filename,
                ):
                    if zip_info.is_dir():
                        (destination / name).mkdir()
                    else:
                        logger.info(
                            "File %s (%s) to %s",
                            zip_info.filename,
                            name,
                            destination / name,
                        )
                        with (
                            zip.open(zip_info) as fp,
                            (destination / name).open("wb") as out,
                        ):
                            shutil.copyfileobj(fp, out)


class TarDownloader(ArchiveDownloader):
    """TAR archive handler."""

    def _name(self, name):
        return re.sub(r"\.tar(\.gz|\.bz\|xz)?$", "", name)

    def unarchive(self, file: CachedFile, destination: Path):
        logger.info("Unarchiving file")
        if self.subpath:
            raise NotImplementedError()

        with tarfile.TarFile.open(file.path) as tar:
            if self.extractall:
                tar.extractall(destination)
            else:
                for info, name in self.filter(tar, lambda info: info.name):
                    if info.isdir():
                        (destination / name).mkdir()
                    else:
                        logger.info(
                            "File %s (%s) to %s",
                            info.name,
                            name,
                            destination / name,
                        )
                        logger.info(
                            "Extracting into %s",
                            destination / name,
                        )
                        tar.extract(info, destination / name)


# Factory aliases for backward compat and convenient usage
zipdownloader = ZipDownloader.apply
tardownloader = TarDownloader.apply



from pathlib import Path
import datamaestro.download.single as single
from datamaestro.definitions import AbstractDataset
from .conftest import MyRepository


TEST_PATH = Path(__file__).parent


class Dataset(AbstractDataset):
    def __init__(self, repository):
        super().__init__(repository)
        self.datapath = Path(repository.context._path)

    def _prepare(self):
        pass


def test_filedownloader(context):
    repository = MyRepository(context)
    dataset = Dataset(repository)

    url = "http://httpbin.org/html"
    downloader = single.filedownloader("test", url)
    downloader(dataset)
    downloader.download()



"""Multiple download resources (legacy).

Note: This module uses a legacy API pattern and needs deeper refactoring.
The List and Datasets classes use an older constructor signature that
differs from the modern Resource interface.
"""

import logging
import os
import warnings
from pathlib import Path

from datamaestro.definitions import AbstractDataset
from datamaestro.download import Download

warnings.warn(
    "datamaestro.download.multiple uses a legacy API. "
    "Consider migrating to class-attribute resource definitions.",
    DeprecationWarning,
    stacklevel=2,
)


class List(Download):
    """Download multiple files or directories given by a list.

    Legacy: uses old-style constructor API.
    """

    def __init__(self, dataset: AbstractDataset, definition: object):
        super().__init__(dataset, definition)
        self.list = self.definition

    def download(self, destination):
        logging.info("Downloading %d items", len(self.list))
        for key, value in self.list.items():
            if not key.startswith("__"):
                handler = Download.find(self.dataset, value)
                destpath = handler.path(destination, key)
                handler.download(destpath)

    def files(self, destpath):
        """Set the list of files"""
        r = {}
        for key, value in self.list.items():
            if not key.startswith("__"):
                handler = Download.find(self.dataset, value)
                r[key] = handler.files(destpath)
        return r


class Datasets(Download):
    """Use links to dataset files.

    Legacy: uses old-style constructor API.
    """

    def __init__(self, dataset: AbstractDataset, definition: object):
        super().__init__(dataset, definition)
        self.list = self.definition

    def download(self, destination):
        destination.mkdir(exist_ok=True, parents=True)

        for key, value in self.list.items():
            if not key.startswith("__"):
                files = value.prepare().files

                if isinstance(files, Path):
                    if not files.is_dir():
                        raise AssertionError(
                            "Dataset path is not a directory: %s",
                            files,
                        )
                    path = destination / key
                    if not path.exists():
                        if path.is_symlink():
                            logging.warning("Path %s is symlink", path)
                        else:
                            os.symlink(files, path)
                elif len(files) > 1:
                    raise NotImplementedError()

    def files(self, destpath):
        """Set the list of files"""
        return destpath
